#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Tests for Ideal + Ideal Liquid (i.e. no activity coefficient) state block;
only tests for construction as parameters need to be provided or estimated
from VLE data to compute the activity coefficients.

Author: Jaffer Ghouse
"""
import pytest
from pyomo.environ import ConcreteModel, TerminationCondition, \
    SolverStatus, value
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock
from idaes.generic_models.properties.activity_coeff_models.BTX_activity_coeff_VLE \
    import BTXParameterBlock
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util import get_solver

solver = get_solver()

# -----------------------------------------------------------------------------
# Create a flowsheet for test
m = ConcreteModel()
m.fs = FlowsheetBlock(default={"dynamic": False})

# vapor-liquid (ideal) - FTPz
m.fs.properties_ideal_vl_FTPz = BTXParameterBlock(
    default={"valid_phase": ('Liq', 'Vap'),
             "activity_coeff_model": "Ideal",
             "state_vars": "FTPz"})
m.fs.state_block_ideal_vl_FTPz = m.fs.properties_ideal_vl_FTPz.build_state_block(
    default={"defined_state": True})

# liquid only (ideal)
m.fs.properties_ideal_l = BTXParameterBlock(default={"valid_phase":
                                                     'Liq',
                                                     "activity_coeff_model":
                                                     "Ideal"})
m.fs.state_block_ideal_l = m.fs.properties_ideal_l.build_state_block(
    default={"has_phase_equilibrium": False,
             "defined_state": True})

# vapour only (ideal)
m.fs.properties_ideal_v = BTXParameterBlock(default={"valid_phase":
                                                     'Vap',
                                                     "activity_coeff_model":
                                                     "Ideal"})
m.fs.state_block_ideal_v = m.fs.properties_ideal_v.build_state_block(
    default={"has_phase_equilibrium": False,
             "defined_state": True})


@pytest.mark.unit
def test_build_inlet_state_block():
    assert len(m.fs.properties_ideal_vl_FTPz.config) == 4

    # vapor-liquid (ideal)
    assert m.fs.properties_ideal_vl_FTPz.config.valid_phase == ('Vap', 'Liq') or \
        m.fs.properties_ideal_vl_FTPz.config.valid_phase == ('Liq', 'Vap')
    assert len(m.fs.properties_ideal_vl_FTPz.phase_list) == 2
    assert m.fs.properties_ideal_vl_FTPz.phase_list == ["Liq", "Vap"]
    assert m.fs.state_block_ideal_vl_FTPz.config.defined_state
    assert hasattr(m.fs.state_block_ideal_vl_FTPz, "eq_phase_equilibrium")
    assert not hasattr(m.fs.state_block_ideal_vl_FTPz, "eq_activity_coeff")
    assert not hasattr(m.fs.state_block_ideal_vl_FTPz, "eq_mol_frac_out")

    # liquid only (ideal)
    assert len(m.fs.properties_ideal_l.config) == 4

    assert m.fs.properties_ideal_l.config.valid_phase == 'Liq'
    assert len(m.fs.properties_ideal_l.phase_list) == 1
    assert m.fs.properties_ideal_l.phase_list == ["Liq"]
    assert m.fs.state_block_ideal_l.config.defined_state
    assert not hasattr(m.fs.state_block_ideal_l, "eq_phase_equilibrium")
    assert not hasattr(m.fs.state_block_ideal_l, "eq_activity_coeff")
    assert not hasattr(m.fs.state_block_ideal_l, "eq_mol_frac_out")

    # vapor only (ideal)
    assert len(m.fs.properties_ideal_v.config) == 4

    assert m.fs.properties_ideal_v.config.valid_phase == 'Vap'
    assert len(m.fs.properties_ideal_v.phase_list) == 1
    assert m.fs.properties_ideal_v.phase_list == ["Vap"]
    assert m.fs.state_block_ideal_v.config.defined_state
    assert not hasattr(m.fs.state_block_ideal_v, "eq_phase_equilibrium")
    assert not hasattr(m.fs.state_block_ideal_v, "eq_activity_coeff")
    assert not hasattr(m.fs.state_block_ideal_v, "eq_mol_frac_out")


@pytest.mark.unit
def test_setInputs_inlet_state_block():

    # vapor-liquid (ideal)
    m.fs.state_block_ideal_vl_FTPz.flow_mol.fix(1)
    m.fs.state_block_ideal_vl_FTPz.temperature.fix(368)
    m.fs.state_block_ideal_vl_FTPz.pressure.fix(101325)
    m.fs.state_block_ideal_vl_FTPz.mole_frac_comp["benzene"].fix(0.5)
    m.fs.state_block_ideal_vl_FTPz.mole_frac_comp["toluene"].fix(0.5)

    assert degrees_of_freedom(m.fs.state_block_ideal_vl_FTPz) == 0

    # liquid only (ideal)
    m.fs.state_block_ideal_l.flow_mol.fix(1)
    m.fs.state_block_ideal_l.temperature.fix(368)
    m.fs.state_block_ideal_l.pressure.fix(101325)
    m.fs.state_block_ideal_l.mole_frac_comp["benzene"].fix(0.5)
    m.fs.state_block_ideal_l.mole_frac_comp["toluene"].fix(0.5)

    assert degrees_of_freedom(m.fs.state_block_ideal_l) == 0

    # vapour only (ideal)
    m.fs.state_block_ideal_v.flow_mol.fix(1)
    m.fs.state_block_ideal_v.temperature.fix(368)
    m.fs.state_block_ideal_v.pressure.fix(101325)
    m.fs.state_block_ideal_v.mole_frac_comp["benzene"].fix(0.5)
    m.fs.state_block_ideal_v.mole_frac_comp["toluene"].fix(0.5)

    assert degrees_of_freedom(m.fs.state_block_ideal_v) == 0


@pytest.mark.unit
def test_solve():
    # vapor-liquid
    m.fs.state_block_ideal_vl_FTPz.initialize()
    results = solver.solve(m.fs.state_block_ideal_vl_FTPz, tee=False)

    # Check for optimal solution
    assert results.solver.termination_condition == TerminationCondition.optimal
    assert results.solver.status == SolverStatus.ok

    # Check for VLE results
    assert value(m.fs.state_block_ideal_vl_FTPz.mole_frac_phase_comp['Liq',
                                                                'benzene']) == \
        pytest.approx(0.4121, abs=1e-3)
    assert value(m.fs.state_block_ideal_vl_FTPz.mole_frac_phase_comp['Vap',
                                                                'benzene']) == \
        pytest.approx(0.6339, abs=1e-3)

    # liquid only
    m.fs.state_block_ideal_l.initialize()
    results = solver.solve(m.fs.state_block_ideal_l, tee=False)

    # Check for optimal solution
    assert results.solver.termination_condition == TerminationCondition.optimal
    assert results.solver.status == SolverStatus.ok

    # Check for results
    assert value(m.fs.state_block_ideal_l.mole_frac_phase_comp['Liq',
                                                          'benzene']) == \
        pytest.approx(0.5, abs=1e-3)
    assert value(m.fs.state_block_ideal_l.mole_frac_phase_comp['Liq',
                                                          'toluene']) == \
        pytest.approx(0.5, abs=1e-3)

    # vapor only
    m.fs.state_block_ideal_v.initialize()
    results = solver.solve(m.fs.state_block_ideal_v, tee=False)

    # Check for optimal solution
    assert results.solver.termination_condition == TerminationCondition.optimal
    assert results.solver.status == SolverStatus.ok

    # Check for results
    assert value(m.fs.state_block_ideal_v.mole_frac_phase_comp['Vap',
                                                          'benzene']) == \
        pytest.approx(0.5, abs=1e-3)
    assert value(m.fs.state_block_ideal_v.mole_frac_phase_comp['Vap',
                                                          'toluene']) == \
        pytest.approx(0.5, abs=1e-3)


# Create a flowsheet object to test outlet state blocks
m.fs1 = FlowsheetBlock(default={"dynamic": False})

# vapor-liquid (ideal)
m.fs1.properties_ideal_vl = BTXParameterBlock(default={"valid_phase":
                                                       ('Liq', 'Vap'),
                                                       "activity_coeff_model":
                                                       "Ideal"})
m.fs1.state_block_ideal_vl = m.fs1.properties_ideal_vl.build_state_block(
    default={"defined_state": False})

# liquid only (ideal)
m.fs1.properties_ideal_l = BTXParameterBlock(default={"valid_phase":
                                                      "Liq",
                                                      "activity_coeff_model":
                                                      "Ideal"})
m.fs1.state_block_ideal_l = m.fs1.properties_ideal_l.build_state_block(
    default={"has_phase_equilibrium": False,
             "defined_state": False})

# vapour only (ideal)
m.fs1.properties_ideal_v = BTXParameterBlock(default={"valid_phase":
                                                      "Vap",
                                                      "activity_coeff_model":
                                                      "Ideal"})
m.fs1.state_block_ideal_v = m.fs1.properties_ideal_v.build_state_block(
    default={"has_phase_equilibrium": False,
             "defined_state": False})


@pytest.mark.unit
def test_build_outlet_state_block():
    assert len(m.fs1.properties_ideal_vl.config) == 4

    # vapor-liquid (ideal)
    assert m.fs1.properties_ideal_vl.config.valid_phase == ('Vap', 'Liq') or \
        m.fs1.properties_ideal_vl.config.valid_phase == ('Liq', 'Vap')
    assert len(m.fs1.properties_ideal_vl.phase_list) == 2
    assert m.fs1.properties_ideal_vl.phase_list == ["Liq", "Vap"]
    assert not m.fs1.state_block_ideal_vl.config.defined_state
    assert hasattr(m.fs1.state_block_ideal_vl, "eq_phase_equilibrium")
    assert not hasattr(m.fs1.state_block_ideal_vl, "eq_activity_coeff")
    assert hasattr(m.fs1.state_block_ideal_vl, "eq_mol_frac_out")

    # liquid only (ideal)
    assert len(m.fs1.properties_ideal_l.config) == 4

    assert m.fs1.properties_ideal_l.config.valid_phase == 'Liq'
    assert len(m.fs1.properties_ideal_l.phase_list) == 1
    assert m.fs1.properties_ideal_l.phase_list == ["Liq"]
    assert not m.fs1.state_block_ideal_l.config.defined_state
    assert not hasattr(m.fs1.state_block_ideal_l, "eq_phase_equilibrium")
    assert not hasattr(m.fs1.state_block_ideal_l, "eq_activity_coeff")
    assert hasattr(m.fs1.state_block_ideal_l, "eq_mol_frac_out")

    # vapour only (ideal)
    assert len(m.fs1.properties_ideal_v.config) == 4

    assert m.fs1.properties_ideal_v.config.valid_phase == 'Vap'
    assert len(m.fs1.properties_ideal_v.phase_list) == 1
    assert m.fs1.properties_ideal_v.phase_list == ["Vap"]
    assert not m.fs1.state_block_ideal_v.config.defined_state
    assert not hasattr(m.fs1.state_block_ideal_v, "eq_phase_equilibrium")
    assert not hasattr(m.fs1.state_block_ideal_v, "eq_activity_coeff")
    assert hasattr(m.fs1.state_block_ideal_v, "eq_mol_frac_out")


@pytest.mark.unit
def test_setInputs_outlet_state_block():

    # vapor-liquid (ideal)
    m.fs1.state_block_ideal_vl.flow_mol.fix(1)
    m.fs1.state_block_ideal_vl.temperature.fix(368)
    m.fs1.state_block_ideal_vl.pressure.fix(101325)
    m.fs1.state_block_ideal_vl.mole_frac_comp["benzene"].fix(0.5)

    assert degrees_of_freedom(m.fs1.state_block_ideal_vl) == 0

    # liquid only (ideal)
    m.fs1.state_block_ideal_l.flow_mol.fix(1)
    m.fs1.state_block_ideal_l.temperature.fix(368)
    m.fs1.state_block_ideal_l.pressure.fix(101325)
    m.fs1.state_block_ideal_l.mole_frac_comp["benzene"].fix(0.5)

    assert degrees_of_freedom(m.fs1.state_block_ideal_l) == 0

    # vapour only (ideal)
    m.fs1.state_block_ideal_v.flow_mol.fix(1)
    m.fs1.state_block_ideal_v.temperature.fix(368)
    m.fs1.state_block_ideal_v.pressure.fix(101325)
    m.fs1.state_block_ideal_v.mole_frac_comp["benzene"].fix(0.5)

    assert degrees_of_freedom(m.fs1.state_block_ideal_v) == 0


@pytest.mark.integration
def test_units_consistent():
    assert_units_consistent(m.fs)
    assert_units_consistent(m.fs1)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
import sys
import os
sys.path.append(os.path.abspath('..')) # current folder is ~/examples
from idaes.apps.uncertainty_propagation.uncertainties import quantify_propagate_uncertainty
import pandas as pd
from idaes.apps.uncertainty_propagation.examples.NRTL_model_scripts import NRTL_model, NRTL_model_opt

if __name__ == "__main__":
    variable_name = ["fs.properties.tau['benzene', 'toluene']", "fs.properties.tau['toluene','benzene']"]
    current_path = os.path.dirname(os.path.realpath(__file__))
    data = pd.read_csv(os.path.join(current_path, 'BT_NRTL_dataset.csv'))
    def SSE(model, data):
        expr = ((float(data["vap_benzene"]) -
                 model.fs.flash.vap_outlet.mole_frac_comp[0, "benzene"])**2 +
                (float(data["liq_benzene"]) -
                 model.fs.flash.liq_outlet.mole_frac_comp[0, "benzene"])**2)
        return expr*1E4

    results =  quantify_propagate_uncertainty(NRTL_model,NRTL_model_opt, data, variable_name, SSE)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Find documentation for modules and classes in the
generated Sphinx documentation and return its location.
"""
# stdlib
import logging
import os
from pathlib import Path
import requests
import types
from urllib.parse import urlparse

# third-party
from lxml import html

__author__ = "Dan Gunter <dkgunter@lbl.gov>"

_log = logging.getLogger(__name__)


def find_html_docs(dmf, obj=None, obj_name=None, **kw):
    """Get one or more files with HTML documentation for
    the given object, in paths referred to by the dmf instance.
    """
    if obj_name is None:
        module_, name = _get_object_name(obj)
    else:
        if obj_name.endswith("."):
            module_, name = obj_name[:-1], ""
        else:
            try:
                module_, name = obj_name.rsplit(".", 1)
            except ValueError:
                raise
    _log.debug("Find docs for object. module=/{}/ name=/{}/".format(module_, name))
    return get_html_docs(dmf, module_, name, **kw)


def get_html_docs(dmf, module_, name, sphinx_version=(1, 5, 5)):
    paths = dmf.get_doc_paths()
    if not paths:
        raise ValueError("No documentation locations configured")

    _log.info(
        "find HTML docs for module={} class={} on paths={}".format(module_, name, paths)
    )
    location = None
    for p in paths:
        _log.debug(f"Examine help path '{p}'")
        parsed = urlparse(p)
        html_content = None
        is_web = parsed.scheme in ("http", "https")
        if is_web:
            _log.debug(f"Get help from online documentation")
            url = p + "/genindex.html"
            _log.debug(f"reading index file: {url}")
            try:
                response = requests.get(url)
                if response.status_code == 200:
                    html_content = response.text
                else:
                    raise requests.RequestException(f"Bad response code: {response.status_code}")
            except requests.RequestException as err:
                _log.debug(f"Error getting documentation index from '{url}': {err}")
        else:
            _log.debug(f"Get help from local documentation files")
            index_path = Path(p) / "genindex.html"
            if index_path.exists():
                _log.debug(f"reading index file: {index_path}")
                html_file = str(index_path)
                html_content = open(html_file).read()
            else:
                _log.debug(f"Index 'genindex.html' not found at path '{p}'")
        if html_content:
            _log.debug("parsing index file")
            tree = html.fromstring(html_content)
            # Look for manual references first
            refs = _find_refs(tree, module_, name, sphinx_version)
            # Get full paths for relative refs
            if refs:
                ref = _pick_best_ref(refs)
                if is_web:
                    parsed_ref = urlparse(ref)
                    location = f"{p}/{parsed_ref.path}"
                else: # file
                    if os.path.isabs(p):
                        location = os.path.join(p, ref)
                    else:
                        location = os.path.join(dmf.root, p, ref)
        if location:
            _log.debug(f"Found documentation at: {location}")
            break  # stop once we find something
    return [location] if location else []


def _get_object_name(obj):
    if isinstance(obj, types.ModuleType):
        module, name = obj.__name__, ""
    else:
        if hasattr(obj, "_orig_module"):
            module = obj._orig_module
        else:
            module = obj.__module__
        if hasattr(obj, "_orig_name"):
            name = obj._orig_name
        elif isinstance(obj, type):
            name = obj.__name__  # class name of a class
        else:
            name = obj.__class__.__name__
    return module, name


def _find_refs(tree, module, name, sphinx_version):
    """Find object/etc. refs in the HTML.
    """
    _log.debug(f"looking for module {module}")
    # There are 3 types of refs we are looking for
    # 1. manually added with index directive
    xpath_expr = '//li[contains(.,"{m}")]/ul/li/a'.format(m=module)
    _log.debug("manually indexed: xpath expr={}".format(xpath_expr))
    elements = tree.xpath(xpath_expr)
    hrefs = [e.get("href") for e in elements if e.text.strip() == name]
    if hrefs:
        _log.debug("found manually indexed docs at: {}".format(hrefs[0]))
        return hrefs  # found something; stop
    if name:
        # 2a. embedded in the page
        target = "{}.{}".format(module, name)
        subtype = "embedded"  # for logging
    else:
        # 2b. generated by autodoc in the apidoc folder
        target = "module-{}".format(module)  # how Sphinx names these
        subtype = "autodoc"  # for logging
    xpath_expr = '//li/a[contains(@href,"#{}")]/@href'.format(target)
    _log.debug("{}: xpath expr={}".format(subtype, xpath_expr))
    hrefs = [p for p in tree.xpath(xpath_expr) if p.endswith(target)]
    if hrefs:
        _log.debug("found {} docs at: {}".format(subtype, hrefs[0]))
    return hrefs


def _pick_best_ref(refs):
    # trivial case:q
    if len(refs) == 1:
        return refs[0]
    # find any non-apidoc ref
    for r in refs:
        if not r.startswith("apidoc"):
            return r
    # otherwise just return first
    return refs[0]



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
General pure component electrolyte methods
"""
from pyomo.environ import Var, units as pyunits

from idaes.core.util.misc import set_param_from_config


# -----------------------------------------------------------------------------
# Method for constant relative permittivity
# RElative permittivity was referred to as the dielectric constant in the past
class relative_permittivity_constant():

    @staticmethod
    def build_parameters(cobj):
        cobj.relative_permittivity_liq_comp = Var(
            doc="Relative permittivity",
            units=pyunits.dimensionless)
        set_param_from_config(cobj, param="relative_permittivity_liq_comp")

    @staticmethod
    def return_expression(b, cobj, T):
        return cobj.relative_permittivity_liq_comp



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################

import time

_command_import_start_time = time.time()

from idaes.commands.base import command_base as cb

import pkgutil
# import all the commands
for loader, module_name, is_pkg in pkgutil.walk_packages(__path__):
    if module_name == "base":
        pass
    elif module_name.startswith("test"):
        pass
    else:
        loader.find_module(module_name).load_module(module_name)

_command_import_total_time = time.time() - _command_import_start_time



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Tests for methods from electrolyte library

Authors: Andrew Lee
"""

import pytest
import types

from pyomo.environ import \
    ConcreteModel, Block, Expression, value, Var, units as pyunits
from pyomo.common.config import ConfigBlock
from pyomo.util.check_units import assert_units_equivalent

from idaes.generic_models.properties.core.pure.electrolyte import \
    relative_permittivity_constant
from idaes.core.util.misc import add_object_reference
from idaes.core.property_meta import PropertyClassMetadata


@pytest.fixture()
def frame():
    m = ConcreteModel()

    # Create a dummy parameter block
    m.params = Block()

    m.params.config = ConfigBlock(implicit=True)
    m.params.config.parameter_data = {
        "relative_permittivity_liq_comp": 101}

    m.meta_object = PropertyClassMetadata()
    m.meta_object.default_units["temperature"] = pyunits.K
    m.meta_object.default_units["mass"] = pyunits.kg
    m.meta_object.default_units["length"] = pyunits.m
    m.meta_object.default_units["time"] = pyunits.s
    m.meta_object.default_units["amount"] = pyunits.mol

    def get_metadata(self):
        return m.meta_object
    m.get_metadata = types.MethodType(get_metadata, m)
    m.params.get_metadata = types.MethodType(get_metadata, m.params)

    # Add necessary parameters to parameter block
    m.params.temperature_ref = Var(initialize=298.15, units=pyunits.K)
    m.params.pressure_ref = Var(initialize=1e5, units=pyunits.Pa)

    # Create a dummy state block
    m.props = Block([1])
    add_object_reference(m.props[1], "params", m.params)

    m.props[1].temperature = Var(initialize=500, units=pyunits.K)
    m.props[1].pressure = Var(initialize=101325, units=pyunits.Pa)

    return m


@pytest.mark.unit
def test_relative_permitivity_constant(frame):
    relative_permittivity_constant.build_parameters(frame.params)

    assert isinstance(frame.params.relative_permittivity_liq_comp, Var)
    assert value(frame.params.relative_permittivity_liq_comp) == 101

    expr = relative_permittivity_constant.return_expression(
        frame.props[1], frame.params, frame.props[1].temperature)
    assert value(expr) == 101

    frame.props[1].temperature.value = 600
    assert value(expr) == 101

    assert_units_equivalent(expr, pyunits.dimensionless)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Tests for ideal equation of state methods

Author: Andrew Lee
"""
import pytest
from sys import modules

from pyomo.environ import (
    ConcreteModel, Expression, Var, units as pyunits, value)
from pyomo.util.check_units import assert_units_equivalent

from idaes.core import (declare_process_block_class,
                        LiquidPhase, VaporPhase, SolidPhase,
                        Solute, Solvent, Apparent)
from idaes.generic_models.properties.core.eos.ideal import Ideal
from idaes.generic_models.properties.core.generic.generic_property import (
        GenericParameterData)
from idaes.core.util.exceptions import (
    ConfigurationError, PropertyNotSupportedError)
from idaes.core.util.constants import Constants as const


# Dummy method for property method calls
def dummy_call(b, j, T):
    return 42


def dummy_call2(b, j, T):
    return 7


# Dummy method to avoid errors when setting metadata dict
def set_metadata(b):
    pass


@declare_process_block_class("DummyParameterBlock")
class DummyParameterData(GenericParameterData):
    def configure(self):
        self.configured = True

    def parameters(self):
        self.parameters_set = True


def define_state(b):
    b.state_defined = True


@pytest.fixture()
def m():
    m = ConcreteModel()

    # Dummy params block
    m.params = DummyParameterBlock(default={
                "components": {"a": {}, "b": {}, "c": {}},
                "phases": {
                    "Vap": {"type": VaporPhase,
                            "equation_of_state": Ideal},
                    "Liq": {"type": LiquidPhase,
                            "equation_of_state": Ideal}},
                "base_units": {"time": pyunits.s,
                               "length": pyunits.m,
                               "mass": pyunits.kg,
                               "amount": pyunits.mol,
                               "temperature": pyunits.K},
                "state_definition": modules[__name__],
                "pressure_ref": 1e5,
                "temperature_ref": 300})

    m.props = m.params.state_block_class([1],
                                         default={"defined_state": False,
                                         "parameters": m.params})

    # Add common variables
    m.props[1].pressure = Var(initialize=101325)
    m.props[1].temperature = Var(initialize=300, units=pyunits.K)
    m.props[1]._teq = Var([("Vap", "Liq")], initialize=300)
    m.props[1].mole_frac_phase_comp = Var(m.params.phase_list,
                                          m.params.component_list,
                                          initialize=0.5)

    return m


@pytest.fixture()
def m_sol():
    m = ConcreteModel()

    # Dummy params block with a Solid Phase to check phase typing
    m.params = DummyParameterBlock(default={
                "components": {"a": {}, "b": {}, "c": {}},
                "phases": {
                    "Sol": {"type": SolidPhase,
                            "equation_of_state": Ideal},
                    "Liq": {"type": LiquidPhase,
                            "equation_of_state": Ideal}},
                "base_units": {"time": pyunits.s,
                               "length": pyunits.m,
                               "mass": pyunits.kg,
                               "amount": pyunits.mol,
                               "temperature": pyunits.K},
                "state_definition": modules[__name__],
                "pressure_ref": 1e5,
                "temperature_ref": 300})

    m.props = m.params.build_state_block([1],
                                         default={"defined_state": False})

    # Add common variables
    m.props[1].pressure = Var(initialize=101325)
    m.props[1].temperature = Var(initialize=300)
    m.props[1]._teq = Var([("Vap", "Liq")], initialize=300)
    m.props[1].mole_frac_phase_comp = Var(m.params.phase_list,
                                          m.params.component_list,
                                          initialize=0.5)

    return m


@pytest.mark.unit
def test_common(m):
    assert Ideal.common(m.props, "foo") is None


@pytest.mark.unit
def test_compress_fact_phase_Liq(m):
    assert Ideal.compress_fact_phase(m.props[1], "Liq") == 0


@pytest.mark.unit
def test_compress_fact_phase_Vap(m):
    assert Ideal.compress_fact_phase(m.props[1], "Vap") == 1


@pytest.mark.unit
def test_compress_fact_phase_sol(m_sol):
    assert Ideal.compress_fact_phase(m_sol.props[1], "Sol") == 0


@pytest.mark.unit
def test_cp_mol_phase(m):
    m.props[1].cp_mol_phase_comp = Var(m.params.phase_list,
                                       m.params.component_list)

    for p in m.params.phase_list:
        assert str(Ideal.cp_mol_phase(m.props[1], p)) == str(
            sum(m.props[1].mole_frac_phase_comp[p, j] *
                m.props[1].cp_mol_phase_comp[p, j]
                for j in m.params.component_list))


@pytest.mark.unit
def test_cp_mol_phase_comp(m):
    for j in m.params.component_list:
        m.params.get_component(j).config.cp_mol_liq_comp = dummy_call
        m.params.get_component(j).config.cp_mol_ig_comp = dummy_call

        assert str(Ideal.cp_mol_phase_comp(m.props[1], "Liq", j)) == str(42)
        assert str(Ideal.cp_mol_phase_comp(m.props[1], "Vap", j)) == str(42)


@pytest.mark.unit
def test_cv_mol_phase(m):
    m.props[1].cv_mol_phase_comp = Var(m.params.phase_list,
                                       m.params.component_list)

    for p in m.params.phase_list:
        assert str(Ideal.cv_mol_phase(m.props[1], p)) == str(
            sum(m.props[1].mole_frac_phase_comp[p, j] *
                m.props[1].cv_mol_phase_comp[p, j]
                for j in m.params.component_list))


@pytest.mark.unit
def test_cv_mol_phase_comp(m):
    for j in m.params.component_list:
        m.params.get_component(j).config.cp_mol_liq_comp = dummy_call
        m.params.get_component(j).config.cp_mol_ig_comp = dummy_call

        assert str(Ideal.cv_mol_phase_comp(m.props[1], "Liq", j)) == str(42)
        assert str(Ideal.cv_mol_phase_comp(m.props[1], "Vap", j)) == str(
            42 - pyunits.convert(
                const.gas_constant,
                to_units=pyunits.kg*pyunits.m**2/pyunits.s**2 /
                pyunits.mol/pyunits.K))


@pytest.mark.unit
def test_heat_capacity_ratio_phase(m):

    m.props[1].cp_mol_phase = Var(m.params.phase_list)
    m.props[1].cv_mol_phase = Var(m.params.phase_list)

    for p in m.params.phase_list:
        assert str(Ideal.heat_capacity_ratio_phase(m.props[1], p)) == (
            str(m.props[1].cp_mol_phase[p] / m.props[1].cv_mol_phase[p]))


@pytest.mark.unit
def test_dens_mass_phase(m):
    m.props[1].dens_mol_phase = Var(m.params.phase_list)
    m.props[1].mw_phase = Var(m.params.phase_list)

    for p in m.params.phase_list:
        assert str(Ideal.dens_mass_phase(m.props[1], p)) == str(
                m.props[1].dens_mol_phase[p]*m.props[1].mw_phase[p])


@pytest.mark.unit
def test_dens_mol_phase_liq(m):
    for j in m.params.component_list:
        m.params.get_component(j).config.dens_mol_liq_comp = dummy_call

    assert str(Ideal.dens_mol_phase(m.props[1], "Liq")) == str(
                1/(1/42*m.props[1].mole_frac_phase_comp["Liq", "a"] +
                   1/42*m.props[1].mole_frac_phase_comp["Liq", "b"] +
                   1/42*m.props[1].mole_frac_phase_comp["Liq", "c"]))


@pytest.mark.unit
def test_dens_mol_phase_vap(m):
    assert str(Ideal.dens_mol_phase(m.props[1], "Vap")) == (
            'props[1].pressure/(kg*m**2/J/s**2*(8.314462618*(J)/mol/K)'
            '*props[1].temperature)')


@pytest.mark.unit
def test_dens_mol_phase_sol(m_sol):
    for j in m_sol.params.component_list:
        m_sol.params.get_component(j).config.dens_mol_sol_comp = dummy_call
        m_sol.params.get_component(j).config.dens_mol_liq_comp = dummy_call

    assert str(Ideal.dens_mol_phase(m_sol.props[1], "Sol")) == str(
        1/sum(m_sol.props[1].mole_frac_phase_comp["Sol", j]/42
              for j in m_sol.params.component_list))


@pytest.mark.unit
def test_energy_internal_mol_phase(m):
    m.props[1].energy_internal_mol_phase_comp = Var(m.params.phase_list,
                                                    m.params.component_list)

    for p in m.params.phase_list:
        assert str(Ideal.energy_internal_mol_phase(m.props[1], p)) == str(
            sum(m.props[1].mole_frac_phase_comp[p, j] *
                m.props[1].energy_internal_mol_phase_comp[p, j]
                for j in m.params.component_list))


@pytest.mark.unit
def test_energy_internal_mol_phase_comp_no_h_form(m):
    for j in m.params.component_list:
        m.params.config.include_enthalpy_of_formation = False
        m.params.get_component(j).config.enth_mol_liq_comp = dummy_call
        m.params.get_component(j).config.enth_mol_ig_comp = dummy_call

        assert (
            str(Ideal.energy_internal_mol_phase_comp(m.props[1], "Liq", j)) ==
            str(42))
        assert (
            str(Ideal.energy_internal_mol_phase_comp(m.props[1], "Vap", j)) ==
            str(42 - pyunits.convert(
                const.gas_constant,
                to_units=pyunits.kg*pyunits.m**2/pyunits.s**2 /
                pyunits.mol/pyunits.K) *
                (m.props[1].temperature - m.params.temperature_ref)))


@pytest.mark.unit
def test_energy_internal_mol_phase_comp_with_h_form(m):
    for j in m.params.component_list:
        m.params.config.include_enthalpy_of_formation = True
        m.params.get_component(j).config.elemental_composition = {
            "C": 1, "He": 2, "O": 4}
        m.params.get_component(j).config.enth_mol_liq_comp = dummy_call
        m.params.get_component(j).config.enth_mol_ig_comp = dummy_call

        # For liquid phase, delta(n) should be 4 (2*He + 2*O2)
        assert (
            str(Ideal.energy_internal_mol_phase_comp(m.props[1], "Liq", j)) ==
            "42 -4.0*(kg*m**2/J/s**2*(8.314462618*(J)/mol/K))"
            "*params.temperature_ref")
        # For vapor phase, delta(n) should be 3 (2*He + 2*O2 - 1*component)
        assert (
            str(Ideal.energy_internal_mol_phase_comp(m.props[1], "Vap", j)) ==
            str(42 - pyunits.convert(
                    const.gas_constant,
                    to_units=pyunits.kg*pyunits.m**2/pyunits.s**2 /
                    pyunits.mol/pyunits.K) *
                (m.props[1].temperature - m.params.temperature_ref) -
                3.0*pyunits.convert(
                    const.gas_constant,
                    to_units=pyunits.kg*pyunits.m**2/pyunits.s**2 /
                    pyunits.mol/pyunits.K) * m.params.temperature_ref))


@pytest.mark.unit
def test_enth_mol_phase(m):
    for j in m.params.component_list:
        m.params.get_component(j).config.enth_mol_liq_comp = dummy_call
        m.params.get_component(j).config.enth_mol_ig_comp = dummy_call
        m.params.get_component(j).config.dens_mol_liq_comp = dummy_call2

    assert str(Ideal.enth_mol_phase(m.props[1], "Vap")) == (
        'props[1].mole_frac_phase_comp[Vap,a]*42.0 + '
        'props[1].mole_frac_phase_comp[Vap,b]*42.0 + '
        'props[1].mole_frac_phase_comp[Vap,c]*42.0')
    assert str(Ideal.enth_mol_phase(m.props[1], "Liq")) == str(
        sum(m.props[1].mole_frac_phase_comp["Liq", j] * 42
            for j in m.params.component_list) +
        (m.props[1].pressure - m.params.pressure_ref) /
        m.props[1].dens_mol_phase["Liq"])


@pytest.mark.unit
def test_enth_mol_phase_comp(m):
    for j in m.params.component_list:
        m.params.get_component(j).config.enth_mol_liq_comp = dummy_call
        m.params.get_component(j).config.enth_mol_ig_comp = dummy_call
        m.params.get_component(j).config.dens_mol_liq_comp = dummy_call2

    for j in m.params.component_list:
        assert str(Ideal.enth_mol_phase_comp(m.props[1], "Liq", j)) == str(
            42 + (m.props[1].pressure - m.params.pressure_ref) /
            m.props[1].dens_mol_phase["Liq"])
        assert str(Ideal.enth_mol_phase_comp(m.props[1], "Vap", j)) == str(42)


@pytest.mark.unit
def test_enth_mol_phase_sol(m_sol):
    for j in m_sol.params.component_list:
        m_sol.params.get_component(j).config.enth_mol_sol_comp = dummy_call
        m_sol.params.get_component(j).config.dens_mol_liq_comp = dummy_call2
        m_sol.params.get_component(j).config.dens_mol_sol_comp = dummy_call2

    for j in m_sol.params.component_list:
        assert str(Ideal.enth_mol_phase_comp(m_sol.props[1], "Sol", j)) == str(
            42 + (m_sol.props[1].pressure - m_sol.params.pressure_ref) /
            m_sol.props[1].dens_mol_phase["Sol"])


@pytest.mark.unit
def test_entr_mol_phase(m):
    m.props[1].entr_mol_phase_comp = Var(m.params.phase_list,
                                         m.params.component_list)

    for p in m.params.phase_list:
        assert str(Ideal.entr_mol_phase(m.props[1], p)) == str(
            sum(m.props[1].mole_frac_phase_comp[p, j] *
                m.props[1].entr_mol_phase_comp[p, j]
                for j in m.params.component_list))


@pytest.mark.unit
def test_entr_mol_phase_comp(m):
    for j in m.params.component_list:
        m.params.get_component(j).config.entr_mol_liq_comp = dummy_call
        m.params.get_component(j).config.entr_mol_ig_comp = dummy_call

        assert str(Ideal.entr_mol_phase_comp(m.props[1], "Liq", j)) == str(42)
        assert str(Ideal.entr_mol_phase_comp(m.props[1], "Vap", j)) == (
            '42 - kg*m**2/J/s**2*(8.314462618*(J)/mol/K)'
            '*log(props[1].mole_frac_phase_comp'
            '[Vap,{}]*props[1].pressure/params.pressure_ref)'.format(j))


@pytest.mark.unit
def test_entr_mol_phase_sol(m_sol):
    for j in m_sol.params.component_list:
        m_sol.params.get_component(j).config.entr_mol_sol_comp = dummy_call

        assert str(Ideal.entr_mol_phase_comp(m_sol.props[1], "Sol", j)) == str(
            42)


@pytest.mark.unit
def test_fug_phase_comp_liq(m):
    for j in m.params.component_list:
        m.params.get_component(j).config.pressure_sat_comp = dummy_call

        assert (str(Ideal.fug_phase_comp(m.props[1], "Liq", j)) ==
                str(m.props[1].mole_frac_phase_comp["Liq", j] * 42))


@pytest.mark.unit
def test_fug_phase_comp_vap(m):
    for j in m.params.component_list:
        assert (str(Ideal.fug_phase_comp(m.props[1], "Vap", j)) ==
                str(m.props[1].mole_frac_phase_comp["Vap", j] *
                    m.props[1].pressure))


@pytest.mark.unit
def test_fug_phase_comp_invalid_phase(m_sol):
    with pytest.raises(PropertyNotSupportedError):
        Ideal.fug_phase_comp(m_sol.props[1], "Sol", "a")


@pytest.mark.unit
def test_fug_phase_comp_liq_eq(m):
    for j in m.params.component_list:
        m.params.get_component(j).config.pressure_sat_comp = dummy_call

        assert (str(Ideal.fug_phase_comp_eq(
                        m.props[1], "Liq", j, ("Vap", "Liq"))) ==
                str(m.props[1].mole_frac_phase_comp["Liq", j] * 42))


@pytest.mark.unit
def test_fug_phase_comp_vap_eq(m):
    for j in m.params.component_list:
        assert (str(Ideal.fug_phase_comp_eq(
                        m.props[1], "Vap", j, ("Vap", "Liq"))) ==
                str(m.props[1].mole_frac_phase_comp["Vap", j] *
                    m.props[1].pressure))


@pytest.mark.unit
def test_fug_phase_comp_invalid_phase_eq(m_sol):
    with pytest.raises(PropertyNotSupportedError):
        Ideal.fug_phase_comp_eq(m_sol.props[1], "Sol", "a", ("Vap", "Liq"))


@pytest.mark.unit
def test_fug_coeff_phase_comp(m):
    for p in m.params.phase_list:
        for j in m.params.component_list:
            assert Ideal.fug_coeff_phase_comp(m.props[1], p, j) == 1


@pytest.mark.unit
def test_fug_coeff_phase_comp_invalid_phase(m_sol):
    with pytest.raises(PropertyNotSupportedError):
        Ideal.fug_coeff_phase_comp(m_sol.props[1], "Sol", "foo")


@pytest.mark.unit
def test_gibbs_mol_phase(m):
    m.props[1].gibbs_mol_phase_comp = Var(m.params.phase_list,
                                          m.params.component_list)

    for p in m.params.phase_list:
        assert str(Ideal.gibbs_mol_phase(m.props[1], p)) == str(
            sum(m.props[1].mole_frac_phase_comp[p, j] *
                m.props[1].gibbs_mol_phase_comp[p, j]
                for j in m.params.component_list))


@pytest.mark.unit
def test_gibbs_mol_phase_comp(m):
    m.props[1].enth_mol_phase_comp = Var(m.params.phase_list,
                                         m.params.component_list)
    m.props[1].entr_mol_phase_comp = Var(m.params.phase_list,
                                         m.params.component_list)

    for p in m.params.phase_list:
        for j in m.params.component_list:
            assert str(Ideal.gibbs_mol_phase_comp(m.props[1], p, j)) == str(
                    m.props[1].enth_mol_phase_comp[p, j] -
                    m.props[1].entr_mol_phase_comp[p, j] *
                    m.props[1].temperature)


@pytest.mark.unit
def test_pressure_osm_phase_no_solvent(m):
    with pytest.raises(
            ConfigurationError,
            match="called for pressure_osm, but no solvents were "
            "defined. Osmotic pressure requires at least one component to be "
            "declared as a solvent."):
        m.props[1].pressure_osm_phase.display()


@pytest.mark.unit
def test_pressure_osm_phase(m):
    # Create a dummy solvent_set
    m.params.solvent_set = ["a"]
    m.props[1].dens_mol_phase = Var(
        m.params.phase_list, initialize=55e3, units=pyunits.mol/pyunits.m**3)

    assert_units_equivalent(m.props[1].pressure_osm_phase["Liq"],
                            pyunits.Pa)
    assert pytest.approx(value(const.gas_constant*300*55e3),
                         rel=1e-6) == value(
        m.props[1].pressure_osm_phase["Liq"])
    assert str(m.props[1].pressure_osm_phase["Liq"]._expr) == (
        'kg*m**2/J/s**2*(' +
        str(const.gas_constant)+')*' +
        str(m.props[1].temperature*(
            m.props[1].conc_mol_phase_comp["Liq", "b"] +
            m.props[1].conc_mol_phase_comp["Liq", "c"])))
    assert len(m.props[1].pressure_osm_phase) == 1


@pytest.mark.unit
def test_pressure_osm_phase_w_apparent_component():
    m = ConcreteModel()

    # Dummy params block
    m.params = DummyParameterBlock(default={
                "components": {"a": {"type": Solvent},
                               "b": {"type": Solute},
                               "c": {"type": Apparent,
                                     "dissociation_species": {"foo": 2}}},
                "phases": {
                    "Vap": {"type": VaporPhase,
                            "equation_of_state": Ideal},
                    "Liq": {"type": LiquidPhase,
                            "equation_of_state": Ideal}},
                "base_units": {"time": pyunits.s,
                               "length": pyunits.m,
                               "mass": pyunits.kg,
                               "amount": pyunits.mol,
                               "temperature": pyunits.K},
                "state_definition": modules[__name__],
                "pressure_ref": 1e5,
                "temperature_ref": 300})

    m.props = m.params.state_block_class([1],
                                         default={"defined_state": False,
                                                  "parameters": m.params})

    # Add common variables
    m.props[1].pressure = Var(initialize=101325)
    m.props[1].temperature = Var(initialize=300, units=pyunits.K)
    m.props[1]._teq = Var([("Vap", "Liq")], initialize=300)
    m.props[1].mole_frac_phase_comp = Var(m.params.phase_list,
                                          m.params.component_list,
                                          initialize=0.5)

    m.props[1].dens_mol_phase = Var(
        m.params.phase_list, initialize=55e3, units=pyunits.mol/pyunits.m**3)

    assert_units_equivalent(m.props[1].pressure_osm_phase["Liq"],
                            pyunits.Pa)
    # Factor of 1.5 accounts for the fact the c is doubled.
    assert pytest.approx(value(const.gas_constant*300*1.5*55e3),
                         rel=1e-6) == value(
        m.props[1].pressure_osm_phase["Liq"])
    assert str(m.props[1].pressure_osm_phase["Liq"]._expr) == (
        'kg*m**2/J/s**2*(' +
        str(const.gas_constant)+')*' +
        str(m.props[1].temperature*(
            m.props[1].conc_mol_phase_comp["Liq", "b"] +
            2*m.props[1].conc_mol_phase_comp["Liq", "c"])))
    assert len(m.props[1].pressure_osm_phase) == 1


@pytest.mark.unit
def test_vol_mol_phase_no_methods(m):
    with pytest.raises(ConfigurationError,
                       match="does not have a method defined to use when "
                       "calculating molar volume and density for component a "
                       "in phase liq. Each component must define a method for "
                       "either vol_mol_liq_comp or dens_mol_liq_comp."):
        Ideal.vol_mol_phase(m.props[1], "Liq")


@pytest.mark.unit
def test_vol_mol_phase():
    m = ConcreteModel()

    # Dummy params block
    m.params = DummyParameterBlock(default={
                "components": {
                    "a": {"dens_mol_liq_comp": dummy_call},
                    "b": {"dens_mol_liq_comp": dummy_call},
                    "c": {"vol_mol_liq_comp": dummy_call}},
                "phases": {
                    "Vap": {"type": VaporPhase,
                            "equation_of_state": Ideal},
                    "Liq": {"type": LiquidPhase,
                            "equation_of_state": Ideal}},
                "base_units": {"time": pyunits.s,
                               "length": pyunits.m,
                               "mass": pyunits.kg,
                               "amount": pyunits.mol,
                               "temperature": pyunits.K},
                "state_definition": modules[__name__],
                "pressure_ref": 1e5,
                "temperature_ref": 300})

    m.props = m.params.state_block_class([1],
                                         default={"defined_state": False,
                                         "parameters": m.params})

    # Add common variables
    m.props[1].pressure = Var(initialize=101325)
    m.props[1].temperature = Var(initialize=300, units=pyunits.K)
    m.props[1]._teq = Var([("Vap", "Liq")], initialize=300)
    m.props[1].mole_frac_phase_comp = Var(m.params.phase_list,
                                          m.params.component_list,
                                          initialize=0.5)

    for p in m.params.phase_list:
        if p == "Vap":
            assert str(Ideal.vol_mol_phase(m.props[1], p)) == (
                "kg*m**2/J/s**2*(8.314462618*(J)/mol/K)*"
                "props[1].temperature/props[1].pressure")
        else:
            assert str(Ideal.vol_mol_phase(m.props[1], p)) == str(
                1/42*m.props[1].mole_frac_phase_comp["Liq", "a"] +
                1/42*m.props[1].mole_frac_phase_comp["Liq", "b"] +
                42*m.props[1].mole_frac_phase_comp["Liq", "c"])



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Methods for calculating pure component properties from:

Perry's Chemical Engineers' Handbook, 7th Edition
Perry, Green, Maloney, 1997, McGraw-Hill

All parameter indicies and units based on conventions used by the source
"""
from pyomo.environ import log, Var, units as pyunits

from idaes.core.util.misc import set_param_from_config


# -----------------------------------------------------------------------------
# Heat capacities, enthalpies and entropies
class cp_mol_liq_comp():

    @staticmethod
    def build_parameters(cobj):
        cobj.cp_mol_liq_comp_coeff_1 = Var(
            doc="Parameter 1 for liquid phase molar heat capacity",
            units=pyunits.J*pyunits.kmol**-1*pyunits.K**-1)
        set_param_from_config(cobj, param="cp_mol_liq_comp_coeff", index="1")

        cobj.cp_mol_liq_comp_coeff_2 = Var(
            doc="Parameter 2 for liquid phase molar heat capacity",
            units=pyunits.J*pyunits.kmol**-1*pyunits.K**-2)
        set_param_from_config(cobj, param="cp_mol_liq_comp_coeff", index="2")

        cobj.cp_mol_liq_comp_coeff_3 = Var(
            doc="Parameter 3 for liquid phase molar heat capacity",
            units=pyunits.J*pyunits.kmol**-1*pyunits.K**-3)
        set_param_from_config(cobj, param="cp_mol_liq_comp_coeff", index="3")

        cobj.cp_mol_liq_comp_coeff_4 = Var(
            doc="Parameter 4 for liquid phase molar heat capacity",
            units=pyunits.J*pyunits.kmol**-1*pyunits.K**-4)
        set_param_from_config(cobj, param="cp_mol_liq_comp_coeff", index="4")

        cobj.cp_mol_liq_comp_coeff_5 = Var(
            doc="Parameter 5 for liquid phase molar heat capacity",
            units=pyunits.J*pyunits.kmol**-1*pyunits.K**-5)
        set_param_from_config(cobj, param="cp_mol_liq_comp_coeff", index="5")

    @staticmethod
    def return_expression(b, cobj, T):
        # Specific heat capacity
        T = pyunits.convert(T, to_units=pyunits.K)
        cp = (cobj.cp_mol_liq_comp_coeff_5*T**4 +
              cobj.cp_mol_liq_comp_coeff_4*T**3 +
              cobj.cp_mol_liq_comp_coeff_3*T**2 +
              cobj.cp_mol_liq_comp_coeff_2*T +
              cobj.cp_mol_liq_comp_coeff_1)

        units = b.params.get_metadata().derived_units
        return pyunits.convert(cp, units["heat_capacity_mole"])


class enth_mol_liq_comp():

    @staticmethod
    def build_parameters(cobj):
        if not hasattr(cobj, "cp_mol_liq_comp_coeff_1"):
            cp_mol_liq_comp.build_parameters(cobj)

        if cobj.parent_block().config.include_enthalpy_of_formation:
            units = cobj.parent_block().get_metadata().derived_units

            cobj.enth_mol_form_liq_comp_ref = Var(
                    doc="Liquid phase molar heat of formation @ Tref",
                    units=units["energy_mole"])
            set_param_from_config(cobj, param="enth_mol_form_liq_comp_ref")

    @staticmethod
    def return_expression(b, cobj, T):
        # Specific enthalpy
        T = pyunits.convert(T, to_units=pyunits.K)
        Tr = pyunits.convert(b.params.temperature_ref, to_units=pyunits.K)

        units = b.params.get_metadata().derived_units

        h_form = (cobj.enth_mol_form_liq_comp_ref if
                  b.params.config.include_enthalpy_of_formation
                  else 0*units["energy_mole"])

        h = (pyunits.convert(
                (cobj.cp_mol_liq_comp_coeff_5/5)*(T**5-Tr**5) +
                (cobj.cp_mol_liq_comp_coeff_4/4)*(T**4-Tr**4) +
                (cobj.cp_mol_liq_comp_coeff_3/3)*(T**3-Tr**3) +
                (cobj.cp_mol_liq_comp_coeff_2/2)*(T**2-Tr**2) +
                cobj.cp_mol_liq_comp_coeff_1*(T-Tr), units["energy_mole"]) +
             h_form)

        return h


class entr_mol_liq_comp():

    @staticmethod
    def build_parameters(cobj):
        if not hasattr(cobj, "cp_mol_liq_comp_coeff_1"):
            cp_mol_liq_comp.build_parameters(cobj)

        units = cobj.parent_block().get_metadata().derived_units

        cobj.entr_mol_form_liq_comp_ref = Var(
                doc="Liquid phase molar entropy of formation @ Tref",
                units=units["entropy_mole"])
        set_param_from_config(cobj, param="entr_mol_form_liq_comp_ref")

    @staticmethod
    def return_expression(b, cobj, T):
        # Specific entropy
        T = pyunits.convert(T, to_units=pyunits.K)
        Tr = pyunits.convert(b.params.temperature_ref, to_units=pyunits.K)

        units = b.params.get_metadata().derived_units

        s = (pyunits.convert(
                (cobj.cp_mol_liq_comp_coeff_5/4)*(T**4-Tr**4) +
                (cobj.cp_mol_liq_comp_coeff_4/3)*(T**3-Tr**3) +
                (cobj.cp_mol_liq_comp_coeff_3/2)*(T**2-Tr**2) +
                cobj.cp_mol_liq_comp_coeff_2*(T-Tr) +
                cobj.cp_mol_liq_comp_coeff_1*log(T/Tr),
                units["entropy_mole"]) +
             cobj.entr_mol_form_liq_comp_ref)

        return s


# -----------------------------------------------------------------------------
# Densities
class dens_mol_liq_comp():

    @staticmethod
    def build_parameters(cobj):
        cobj.dens_mol_liq_comp_coeff_1 = Var(
                doc="Parameter 1 for liquid phase molar density",
                units=pyunits.kmol*pyunits.m**-3)
        set_param_from_config(cobj, param="dens_mol_liq_comp_coeff", index="1")

        cobj.dens_mol_liq_comp_coeff_2 = Var(
                doc="Parameter 2 for liquid phase molar density",
                units=pyunits.dimensionless)
        set_param_from_config(cobj, param="dens_mol_liq_comp_coeff", index="2")

        cobj.dens_mol_liq_comp_coeff_3 = Var(
                doc="Parameter 3 for liquid phase molar density",
                units=pyunits.K)
        set_param_from_config(cobj, param="dens_mol_liq_comp_coeff", index="3")

        cobj.dens_mol_liq_comp_coeff_4 = Var(
                doc="Parameter 4 for liquid phase molar density",
                units=pyunits.dimensionless)
        set_param_from_config(cobj, param="dens_mol_liq_comp_coeff", index="4")

    @staticmethod
    def return_expression(b, cobj, T):
        # pg. 2-98
        T = pyunits.convert(T, to_units=pyunits.K)

        rho = (cobj.dens_mol_liq_comp_coeff_1 /
               cobj.dens_mol_liq_comp_coeff_2**(
                   1 + (1-T/cobj.dens_mol_liq_comp_coeff_3) **
                   cobj.dens_mol_liq_comp_coeff_4))

        units = b.params.get_metadata().derived_units

        return pyunits.convert(rho, units["density_mole"])


# -----------------------------------------------------------------------------
class Perrys(object):
    cp_mol_liq_comp = cp_mol_liq_comp
    enth_mol_liq_comp = enth_mol_liq_comp
    entr_mol_liq_comp = entr_mol_liq_comp
    dens_mol_liq_comp = dens_mol_liq_comp



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################

"""
This module contains miscellaneous utility functions for use in IDAES models.
"""
import xml.dom.minidom

import pyomo.environ as pyo
from pyomo.core.base.expression import _GeneralExpressionData
from pyomo.core.base.component import ModelComponentFactory
from pyomo.core.base.indexed_component import (
    UnindexedComponent_set, )
from pyomo.core.base.disable_methods import disable_methods
from pyomo.common.config import ConfigBlock

import idaes.logger as idaeslog
import idaes.core.solvers

_log = idaeslog.getLogger(__name__)


# Author: Andrew Lee
def get_solver(solver=None, options=None):
    """
    General method for getting a solver object which defaults to the standard
    IDAES solver (defined in the IDAES configuration).

    Args:
        solver: string name for desired solver. Default=None, use default solver
        options: dict of solver options to use, overwrites any settings
                 provided by IDAES configuration. Default = None, use default
                 solver options.

    Returns:
        A Pyomo solver object
    """
    if solver is None:
        solver = "default"
    solver_obj = idaes.core.solvers.SolverWrapper(solver, register=False)()

    if options is not None:
        solver_obj.options.update(options)

    return solver_obj


# Author: Andrew Lee
def add_object_reference(self, local_name, remote_object):
    """
    Method to create a reference in the local model to a remote Pyomo object.
    This method should only be used where Pyomo Reference objects are not
    suitable (such as for referencing scalar Pyomo objects where the None
    index is undesirable).

    Args:
        local_name : name to use for local reference (str)
        remote_object : object to make a reference to

    Returns:
        None
    """
    try:
        object.__setattr__(self, local_name, remote_object)
    except AttributeError:
        raise AttributeError(
            "{} failed to construct reference to {} - remote "
            "object does not exist.".format(self.name, remote_object)
        )


# Author: Jaffer Ghouse
def extract_data(data_dict):
    """
    General method that returns a rule to extract data from a python
    dictionary. This method allows the param block to have a database for
    a parameter but extract a subset of this data to initialize a Pyomo
    param object.
    """

    def _rule_initialize(m, *args):
        if len(args) > 1:
            return data_dict[args]
        else:
            return data_dict[args[0]]

    return _rule_initialize


# Author: John Eslick
def TagReference(s, description=""):
    """
    Create a Pyomo reference with an added description string attribute to
    describe the reference. The intended use for these references is to create
    a time-indexed reference to variables in a model corresponding to plant
    measurment tags.

    Args:
        s: Pyomo time slice of a variable or expression
        description (str): A description the measurment

    Returns:
        A Pyomo Reference object with an added doc attribute
    """
    r = pyo.Reference(s)
    r.description = description
    return r


# Author John Eslick
def svg_tag(
    tags,
    svg,
    outfile=None,
    idx=None,
    tag_map=None,
    show_tags=False,
    byte_encoding="utf-8",
    tag_format=None,
    tag_format_default="{:.4e}"
):
    """
    Replace text in a SVG with tag values for the model. This works by looking
    for text elements in the SVG with IDs that match the tags or are in tag_map.

    Args:
        tags: A dictionary where the key is the tag and the value is a Pyomo
            Reference.  The reference could be indexed. In typical IDAES
            applications the references would be indexed by time.
        svg: a file pointer or a string continaing svg contents
        outfile: a file name to save the results, if None don't save
        idx: if None not indexed, otherwise an index in the indexing set of the
            reference
        tag_map: dictionary with svg id keys and tag values, to map svg ids to
            tags
        show_tags: Put tag labels of the diagram instead of numbers
        byte_encoding: If svg is given as a byte-array, use this encoding to
            convert it to a string.
        tag_format: A dictionary of formatting strings.  If the formatting
            string is a callable, it should be a function that takes the value
            to display and returns a formatting string.
        tag_format_default: The default formatting if not explicitly by
            tag_format. If the formatting string is a callable, it should be a
            function that takes the value to display and returns a formatting
            string.

    Returns:
        SVG String
    """
    if tag_format is None:
        tag_format = {}

    if isinstance(svg, str):  # assume this is svg content string
        pass
    elif isinstance(svg, bytes):
        svg = svg.decode(byte_encoding)
    elif hasattr(svg, "read"):  # file-like object to svg
        svg = svg.read()
    else:
        raise TypeError("SVG must either be a string or a file-like object")
    # Make tag map here because the tags may not make valid XML IDs if no
    # tag_map provided we'll go ahead and handle XML @ (maybe more in future)
    if tag_map is None:
        tag_map = dict()
        for tag in tags:
            new_tag = tag.replace("@", "_")
            tag_map[new_tag] = tag
    # Search for text in the svg that has an id in tags
    doc = xml.dom.minidom.parseString(svg)
    texts = doc.getElementsByTagName("text")
    for t in texts:
        id = t.attributes["id"].value
        if id in tag_map:
            # if it's multiline change last line
            try:
                tspan = t.getElementsByTagName("tspan")[-1]
            except IndexError:
                _log.warning(f"Text object but no tspan for tag {tag_map[id]}.")
                _log.warning(f"Skipping output for {tag_map[id]}.")
                continue
            try:
                tspan = tspan.childNodes[0]
            except IndexError:
                # No child node means there is a line with no text, so add some.
                tspan.appendChild(doc.createTextNode(""))
                tspan = tspan.childNodes[0]
            try:
                if show_tags:
                    val = tag_map[id]
                elif idx is None:
                    val = pyo.value(tags[tag_map[id]], exception=False)
                else:
                    val = pyo.value(tags[tag_map[id]][idx], exception=False)
            except ZeroDivisionError:
                val = "Divide_by_0"
            tf = tag_format.get(tag_map[id], tag_format_default)
            try:
                if callable(tf): # conditional formatting
                    tspan.nodeValue = tf(val).format(val)
                else:
                    tspan.nodeValue = tf.format(val)
            except ValueError:
                # whatever it is, it doesn't match the format.  Usually this
                # happens when a string is given, but it is using a default
                # number format
                tspan.nodeValue = val

    new_svg = doc.toxml()
    # If outfile is provided save to a file
    if outfile is not None:
        with open(outfile, "w") as f:
            f.write(new_svg)
    # Return the SVG as a string.  This lets you take several passes at adding
    # output without saving and loading files.
    return new_svg


# Author: John Eslick
def copy_port_values(destination=None, source=None, arc=None,
        direction="forward"):
    """
    Moved to idaes.core.util.initialization.propagate_state.
    Leaving redirection function here for deprecation warning.
    """
    _log.warning("DEPRECATED: copy_port_values has been deprecated. "
            "The same functionality can be found in "
            "idaes.core.util.initialization.propagate_state.")
    from idaes.core.util.initialization import propagate_state
    propagate_state(destination=destination, source=source, arc=arc,
            direction=direction)


def set_param_from_config(b, param, config=None, index=None):
    """
    Utility method to set parameter value from a config block. This allows for
    converting units if required. This method directly sets the value of the
    parameter.

    Args:
        b - block on which parameter and config block are defined
        param - name of parameter as str. Used to find param and config arg
        units - units of param object (used if conversion required)
        config - (optional) config block to get parameter data from. If
                unset, assumes b.config.
        index - (optional) used for pure component properties where a single
                property may have multiple parameters associated with it.

    Returns:
        None
    """
    if config is None:
        try:
            config = b.config
        except AttributeError:
            raise AttributeError(
                "{} - set_param_from_config method was not provided with a "
                "config argument, but no default Config block exists. Please "
                "specify the Config block to use via the config argument."
                .format(b.name))

    # Check that config is an instance of a Config Block
    if not isinstance(config, ConfigBlock):
        raise TypeError(
            "{} - set_param_from_config - config argument provided is not an "
            "instance of a Config Block.".format(b.name))

    if index is None:
        try:
            param_obj = getattr(b, param)
        except AttributeError:
            raise AttributeError(
                "{} - set_param_from_config method was provided with param "
                "argument {}, but no attribute of that name exists."
                .format(b.name, param))

        try:
            p_data = config.parameter_data[param]
        except (KeyError, AttributeError):
            raise KeyError(
                "{} - set_param_from_config method was provided with param "
                "argument {}, but the config block does not contain a "
                "value for this parameter.".format(b.name, param))
    else:
        try:
            param_obj = getattr(b, param+"_"+index)
        except AttributeError:
            raise AttributeError(
                "{} - set_param_from_config method was provided with param and"
                " index arguments {} {}, but no attribute with that "
                "combination ({}_{}) exists."
                .format(b.name, param, index, param, index))

        try:
            p_data = config.parameter_data[param][index]
        except (KeyError, AttributeError):
            raise KeyError(
                "{} - set_param_from_config method was provided with param and"
                " index arguments {} {}, but the config block does not contain"
                " a value for this parameter and index."
                .format(b.name, param, index))

    units = param_obj.get_units()

    if isinstance(p_data, tuple):
        # 11 Dec 2020 - There is currently a bug in Pyomo where trying to
        # convert the units of a unitless quantity results in a TypeError.
        # To avoid this, we check here for cases where both the parameter and
        # user provided value are unitless and bypass unit conversion.
        if ((units is None or units is pyo.units.dimensionless) and
                (p_data[1] is None or p_data[1] is pyo.units.dimensionless)):
            param_obj.value = p_data[0]
        else:
            param_obj.value = pyo.units.convert_value(
                p_data[0], from_units=p_data[1], to_units=units)
    else:
        _log.debug("{} no units provided for parameter {} - assuming default "
                   "units".format(b.name, param))
        param_obj.value = p_data


# -----------------------------------------------------------------------------
# Creating a Component derived from Pyomo's Expression to use in cases
# where an Expression could be mistaken for a Var.
# Author: Andrew Lee
class _GeneralVarLikeExpressionData(_GeneralExpressionData):
    """
    An object derived from _GeneralExpressionData which implements methods for
    common APIs on Vars.

    Constructor Arguments:
        expr: The Pyomo expression stored in this expression.

        component: The Expression object that owns this data.

    Public Class Attributes:
        expr: The expression owned by this data.

    Private class attributes:
        _component: The expression component.
    """

    # Define methods for common APIs on Vars in case user mistakes
    # an Expression for a Var
    @property
    def value(self):
        # Overload value so it behaves like a Var
        return pyo.value(self.expr)

    @value.setter
    def value(self, expr):
        # Overload value seter to prevent users changing the expression body
        raise TypeError(
            "%s is an Expression and does not have a value which can be set."
            % (self.name))

    def setlb(self, val=None):
        raise TypeError(
            "%s is an Expression and can not have bounds. "
            "Use an inequality Constraint instead."
            % (self.name))

    def setub(self, val=None):
        raise TypeError(
            "%s is an Expression and can not have bounds. "
            "Use an inequality Constraint instead."
            % (self.name))

    def fix(self, val=None):
        raise TypeError(
            "%s is an Expression and can not be fixed. "
            "Use an equality Constraint instead."
            % (self.name))

    def unfix(self):
        raise TypeError(
            "%s is an Expression and can not be unfixed."
            % (self.name))

@ModelComponentFactory.register(
    "Named expressions that can be used in places of variables.")
class VarLikeExpression(pyo.Expression):
    """
    A shared var-like expression container, which may be defined over a index.

    Constructor Arguments:
        initialize: A Pyomo expression or dictionary of expressions used
        to initialize this object.

        expr: A synonym for initialize.

        rule: A rule function used to initialize this object.
    """

    _ComponentDataClass = _GeneralVarLikeExpressionData
    NoConstraint    = (1000,)
    Skip            = (1000,)

    def __new__(cls, *args, **kwds):
        if cls is not VarLikeExpression:
            return super(VarLikeExpression, cls).__new__(cls)
        if not args or (args[0] is UnindexedComponent_set and len(args) == 1):
            return super(VarLikeExpression, cls).__new__(
                AbstractSimpleVarLikeExpression)
        else:
            return super(VarLikeExpression, cls).__new__(
                IndexedVarLikeExpression)


class SimpleVarLikeExpression(_GeneralVarLikeExpressionData,
                              VarLikeExpression):

    def __init__(self, *args, **kwds):
        _GeneralVarLikeExpressionData.__init__(self, expr=None, component=self)
        VarLikeExpression.__init__(self, *args, **kwds)

    #
    # From Pyomo: Leaving this method for backward compatibility reasons.
    # (probably should be removed)
    # Note: Doesn't seem to work without it
    #
    def add(self, index, expr):
        """Add an expression with a given index."""
        if index is not None:
            raise KeyError(
                "SimpleExpression object '%s' does not accept "
                "index values other than None. Invalid value: %s"
                % (self.name, index))
        if (type(expr) is tuple) and \
           (expr == pyo.Expression.Skip):
            raise ValueError(
                "Expression.Skip can not be assigned "
                "to an Expression that is not indexed: %s"
                % (self.name))
        self.set_value(expr)
        return self


@disable_methods({'set_value', 'is_constant', 'is_fixed', 'expr'})
class AbstractSimpleVarLikeExpression(SimpleVarLikeExpression):
    pass


class IndexedVarLikeExpression(VarLikeExpression):

    #
    # From Pyomo: Leaving this method for backward compatibility reasons
    # Note: It allows adding members outside of self._index.
    #       This has always been the case. Not sure there is
    #       any reason to maintain a reference to a separate
    #       index set if we allow this.
    #
    def add(self, index, expr):
        """Add an expression with a given index."""
        if (type(expr) is tuple) and (expr == pyo.Expression.Skip):
            return None
        cdata = _GeneralVarLikeExpressionData(expr, component=self)
        self._data[index] = cdata
        return cdata

    # Define methods for common APIs on Vars in case user mistakes
    # an Expression for a Var
    def setlb(self, val=None):
        raise TypeError(
            "%s is an Expression and can not have bounds. "
            "Use inequality Constraints instead."
            % (self.name))

    def setub(self, val=None):
        raise TypeError(
            "%s is an Expression and can not have bounds. "
            "Use inequality Constraints instead."
            % (self.name))

    def fix(self, val=None):
        raise TypeError(
            "%s is an Expression and can not be fixed. "
            "Use equality Constraints instead."
            % (self.name))

    def unfix(self):
        raise TypeError(
            "%s is an Expression and can not be unfixed."
            % (self.name))



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""Tests that helmholtz specific NTU condenser model."""


import pytest
import pyomo.environ as pyo
import idaes.core
from idaes.power_generation.unit_models.helm import HelmNtuCondenser
from idaes.generic_models.properties import iapws95
from idaes.core.util import get_solver

solver = get_solver()

@pytest.mark.unit
@pytest.mark.skipif(solver is None, reason="Solver not available")
def test_condenser_steady():
    m = pyo.ConcreteModel()
    m.fs = idaes.core.FlowsheetBlock(default={"dynamic": False})
    m.fs.properties = iapws95.Iapws95ParameterBlock()
    m.fs.unit = HelmNtuCondenser(
        default={
            "dynamic": False,
            "shell": {
                "has_pressure_change": False,
                "property_package": m.fs.properties,
            },
            "tube": {
                "has_pressure_change": False,
                "property_package": m.fs.properties,
            },
        }
    )

    m.fs.unit.shell_inlet.flow_mol.fix(100)
    m.fs.unit.shell_inlet.pressure[:] = 9000
    m.fs.unit.shell_inlet.enth_mol.fix(30000)

    m.fs.unit.tube_inlet.flow_mol.fix(1200)
    m.fs.unit.tube_inlet.pressure.fix(101325)
    m.fs.unit.tube_inlet.enth_mol.fix(1800)

    m.fs.unit.area.fix(1000)
    m.fs.unit.overall_heat_transfer_coefficient.fix(500)

    m.fs.unit.initialize(unfix='pressure')
    solver.solve(m)

    assert (pyo.value(m.fs.unit.shell_inlet.pressure[0]) ==
        pytest.approx(14036.3360, rel=1e-2))


@pytest.mark.unit
@pytest.mark.skipif(solver is None, reason="Solver not available")
def test_condenser_dynamic():
    m = pyo.ConcreteModel()
    m.fs = idaes.core.FlowsheetBlock(
        default={"dynamic": True, "time_set":[0,3]})
    m.fs.properties = iapws95.Iapws95ParameterBlock()
    m.fs.unit = HelmNtuCondenser(
        default={
            "dynamic": False,
            "shell": {
                "has_pressure_change": False,
                "property_package": m.fs.properties,
            },
            "tube": {
                "has_pressure_change": False,
                "property_package": m.fs.properties,
            },
        }
    )

    pyo.TransformationFactory('dae.finite_difference').apply_to(
        m.fs, nfe=4, wrt=m.fs.time, scheme='BACKWARD')

    m.fs.unit.shell_inlet.flow_mol.fix(100)
    m.fs.unit.shell_inlet.pressure[:] = 9000
    m.fs.unit.shell_inlet.enth_mol.fix(30000)

    m.fs.unit.tube_inlet.flow_mol.fix(1200)
    m.fs.unit.tube_inlet.pressure.fix(101325)
    m.fs.unit.tube_inlet.enth_mol.fix(1800)

    m.fs.unit.area.fix(1000)
    m.fs.unit.overall_heat_transfer_coefficient.fix(500)

    m.fs.unit.initialize(unfix='pressure')
    solver.solve(m)

    for t in m.fs.unit.shell_inlet.pressure:
        assert (pyo.value(m.fs.unit.shell_inlet.pressure[t]) ==
            pytest.approx(14036.3360, rel=1e-2))



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Tests for methods from Perry's

All methods and parameters from:

Perry's Chemical Engineers' Handbook, 7th Edition
Perry, Green, Maloney, 1997, McGraw-Hill

All parameter indicies based on conventions used by the source

Authors: Andrew Lee
"""

import pytest
import types

from pyomo.environ import ConcreteModel, Block, value, Var, units as pyunits
from pyomo.common.config import ConfigBlock
from pyomo.util.check_units import assert_units_equivalent

from idaes.generic_models.properties.core.pure.Perrys import *
from idaes.core.util.misc import add_object_reference
from idaes.core.property_meta import PropertyClassMetadata


@pytest.fixture()
def frame():
    m = ConcreteModel()

    # Create a dummy parameter block
    m.params = Block()

    m.params.config = ConfigBlock(implicit=True)
    m.params.config.parameter_data = {
        "dens_mol_liq_comp_coeff": {'1': 5.459,
                                    '2': 0.30542,
                                    '3': 647.13,
                                    '4': 0.081},
        "cp_mol_liq_comp_coeff": {'1': 2.7637e+05,
                                  '2': -2.0901e+03,
                                  '3': 8.1250e+00,
                                  '4': -1.4116e-2,
                                  '5': 9.3701e-06},
        "enth_mol_form_liq_comp_ref": -285.83e3,
        "entr_mol_form_liq_comp_ref": 69.95}
    m.params.config.include_enthalpy_of_formation = True

    # Also need to dummy configblock on the model for the test
    m.config = ConfigBlock(implicit=True)
    m.config.include_enthalpy_of_formation = True

    m.meta_object = PropertyClassMetadata()
    m.meta_object.default_units["temperature"] = pyunits.K
    m.meta_object.default_units["mass"] = pyunits.kg
    m.meta_object.default_units["length"] = pyunits.m
    m.meta_object.default_units["time"] = pyunits.s
    m.meta_object.default_units["amount"] = pyunits.mol

    def get_metadata(self):
        return m.meta_object
    m.get_metadata = types.MethodType(get_metadata, m)
    m.params.get_metadata = types.MethodType(get_metadata, m.params)

    # Add necessary parameters to parameter block
    m.params.temperature_ref = Var(initialize=273.16, units=pyunits.K)

    # Create a dummy state block
    m.props = Block([1])
    add_object_reference(m.props[1], "params", m.params)

    m.props[1].temperature = Var(initialize=273.16, units=pyunits.K)

    return m


@pytest.mark.unit
def test_cp_mol_liq_comp(frame):
    cp_mol_liq_comp.build_parameters(frame.params)

    assert isinstance(frame.params.cp_mol_liq_comp_coeff_1, Var)
    assert value(frame.params.cp_mol_liq_comp_coeff_1) == 2.7637e+05
    assert isinstance(frame.params.cp_mol_liq_comp_coeff_2, Var)
    assert value(frame.params.cp_mol_liq_comp_coeff_2) == -2.0901e+03
    assert isinstance(frame.params.cp_mol_liq_comp_coeff_3, Var)
    assert value(frame.params.cp_mol_liq_comp_coeff_3) == 8.1250e+00
    assert isinstance(frame.params.cp_mol_liq_comp_coeff_4, Var)
    assert value(frame.params.cp_mol_liq_comp_coeff_4) == -1.4116e-2
    assert isinstance(frame.params.cp_mol_liq_comp_coeff_5, Var)
    assert value(frame.params.cp_mol_liq_comp_coeff_5) == 9.3701e-06

    expr = cp_mol_liq_comp.return_expression(
        frame.props[1], frame.params, frame.props[1].temperature)
    assert value(expr) == pytest.approx(76.150, rel=1e-3)

    frame.props[1].temperature.value = 533.15
    assert value(expr) == pytest.approx(89.390, rel=1e-3)

    assert_units_equivalent(expr, pyunits.J/pyunits.mol/pyunits.K)


@pytest.mark.unit
def test_enth_mol_liq_comp(frame):
    enth_mol_liq_comp.build_parameters(frame.params)

    assert isinstance(frame.params.enth_mol_form_liq_comp_ref, Var)
    assert value(frame.params.enth_mol_form_liq_comp_ref) == -285.83e3

    expr = enth_mol_liq_comp.return_expression(
        frame.props[1], frame.params, frame.props[1].temperature)
    assert value(expr) == value(
            frame.params.enth_mol_form_liq_comp_ref)

    frame.props[1].temperature.value = 533.15
    assert value(expr) == pytest.approx(-265423, rel=1e-3)

    assert_units_equivalent(expr, pyunits.J/pyunits.mol)


@pytest.mark.unit
def test_enth_mol_liq_comp_no_formation(frame):
    frame.config.include_enthalpy_of_formation = False
    frame.params.config.include_enthalpy_of_formation = False

    enth_mol_liq_comp.build_parameters(frame.params)

    assert not hasattr(frame.params, "enth_mol_form_liq_comp_ref")

    expr = enth_mol_liq_comp.return_expression(
        frame.props[1], frame.params, frame.props[1].temperature)
    assert value(expr) == 0

    frame.props[1].temperature.value = 533.15
    assert value(expr) == pytest.approx(-265423 + 285830, rel=1e-3)

    assert_units_equivalent(expr, pyunits.J/pyunits.mol)


@pytest.mark.unit
def test_entr_mol_liq_comp(frame):
    entr_mol_liq_comp.build_parameters(frame.params)

    assert isinstance(frame.params.entr_mol_form_liq_comp_ref, Var)
    assert value(frame.params.entr_mol_form_liq_comp_ref) == 69.95

    expr = entr_mol_liq_comp.return_expression(
        frame.props[1], frame.params, frame.props[1].temperature)
    assert value(expr) == value(
            frame.params.entr_mol_form_liq_comp_ref)

    frame.props[1].temperature.value = 533.15
    assert value(expr) == pytest.approx(122.05, rel=1e-3)

    assert_units_equivalent(expr, pyunits.J/pyunits.mol/pyunits.K)


@pytest.mark.unit
def test_dens_mol_liq_comp(frame):
    dens_mol_liq_comp.build_parameters(frame.params)

    assert isinstance(frame.params.dens_mol_liq_comp_coeff_1, Var)
    assert value(frame.params.dens_mol_liq_comp_coeff_1) == 5.459
    assert isinstance(frame.params.dens_mol_liq_comp_coeff_2, Var)
    assert value(frame.params.dens_mol_liq_comp_coeff_2) == 0.30542
    assert isinstance(frame.params.dens_mol_liq_comp_coeff_3, Var)
    assert value(frame.params.dens_mol_liq_comp_coeff_3) == 647.13
    assert isinstance(frame.params.dens_mol_liq_comp_coeff_4, Var)
    assert value(frame.params.dens_mol_liq_comp_coeff_4) == 0.081

    expr = dens_mol_liq_comp.return_expression(
        frame.props[1], frame.params, frame.props[1].temperature)
    assert value(expr) == pytest.approx(55.583e3, rel=1e-4)

    frame.props[1].temperature.value = 333.15
    assert value(expr) == pytest.approx(54.703e3, rel=1e-4)

    assert_units_equivalent(expr, pyunits.mol/pyunits.m**3)



# coding: utf-8
#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""__init__.py for idaes module

Set up logging for the idaes module, and import plugins.
"""
import os
import copy

from . import config
import logging

from .ver import __version__  # noqa

_log = logging.getLogger(__name__)

# Standard locations for config file, binary libraries and executables, ...
data_directory, bin_directory, testing_directory = config.get_data_directory()
# To avoid a circular import the config module doesn't import idaes, but
# some functions in the config module that are executed later use this
# these directories are static from here on.
config.data_directory = data_directory
config.bin_directory = bin_directory
config.testing_directory = testing_directory

# Set the path for the global and local config files
_global_config_file = os.path.join(data_directory, "idaes.conf")
_local_config_file = "idaes.conf"

# Create the general IDAES configuration block, with default config
cfg = config._new_idaes_config_block()
config.reconfig(cfg)
# read global config and overwrite provided config options
config.read_config(_global_config_file, cfg=cfg)
# read local config and overwrite provided config options
config.read_config(_local_config_file, cfg=cfg)

# Setup the environment so solver executables can be run
config.setup_environment(bin_directory, cfg.use_idaes_solvers)

# Debug log for basic testing of the logging config
_log.debug("'idaes' logger debug test")

def _create_data_dir():
    """Create the IDAES directory to store data files in."""
    config.create_dir(data_directory)

def _create_bin_dir(bd=None):
    """Create the IDAES directory to store executable files in.

    Args:
        bd: alternate binary directory, used for testing
    """
    _create_data_dir()
    if bd is None:
        bd = bin_directory
    config.create_dir(bd)

def _create_testing_dir():
    """Create an idaes testing directory
    """
    _create_data_dir()
    config.create_dir(testing_directory)

try:
    _create_data_dir()
except FileNotFoundError:
    pass # the standard place for this doesn't exist, shouldn't be a show stopper

try:
    _create_bin_dir()
except FileNotFoundError:
    pass # the standard place for this doesn't exist, shouldn't be a show stopper

try:
    _create_testing_dir()
except FileNotFoundError:
    pass # the standard place for this doesn't exist, shouldn't be a show stopper


def reconfig():
    return config.reconfig(cfg)

def read_config(val):
    return config.read_config(val=val, cfg=cfg)

def write_config(path, default=False):
    _cfg = None if default else cfg
    return config.write_config(path=path, cfg=_cfg)


class temporary_config_ctx(object):
    def __enter__(self):
        self.orig_config = copy.deepcopy(cfg)
    def __exit__(self, exc_type, exc_value, traceback):
        global cfg
        cfg = self.orig_config
        reconfig()



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Tests for idaes.dmf.getver module
"""

import idaes
from idaes import dmf
from idaes.dmf import getver
import uuid

import pytest


@pytest.fixture
def garbage():
    return uuid.uuid4().hex


@pytest.fixture
def make_pip_garbage():
    garbage = uuid.uuid4().hex
    orig, getver.Versioned.PIP = getver.Versioned.PIP, garbage
    yield garbage
    getver.Versioned.PIP = orig


@pytest.mark.unit
def test_clazz():
    v1 = getver.Versioned("idaes")
    v2 = getver.Versioned(idaes)
    assert v1.get_info() == v2.get_info()
    assert v1.name == v2.name


@pytest.mark.unit
def test_func():
    import numpy

    i1 = dmf.get_version_info(idaes)
    i2 = dmf.get_version_info(numpy)
    i3 = dmf.get_version_info("pytest")
    i4 = dmf.get_version_info("traitlets")
    assert i1 != i2 and i2 != i3
    # one type of git hash error
    with pytest.raises(getver.GitHashError):
        _ = getver.Versioned("pytest").git_hash
    # another type of git hash error
    with pytest.raises(getver.GitHashError):
        _ = getver.Versioned("traitlets").git_hash


class fake_module:
    """Solely to exercise a path where the loader has no path attr
    """
    __package__ = "pytest"  # needs to match something real
    class fake_loader:
        pass
    __loader__ = fake_loader


@pytest.mark.unit
def test_more_func():
    import six

    # doesn't have __path__ but has __loader__ with path
    dmf.get_version_info(six)
    # __loader__ without path
    dmf.get_version_info(fake_module)


@pytest.mark.unit
def test_bad_import(garbage):
    with pytest.raises(getver.ModuleImportError):
        getver.Versioned(garbage)


@pytest.mark.unit
def test_bad_pip(make_pip_garbage):
    ver = getver.Versioned("traitlets")
    assert ver is not None


@pytest.mark.unit
def test_repeated_calls():
    import time

    t0 = time.time()
    vv = getver.Versioned("idaes")
    i1 = vv.get_info()
    t1 = time.time()
    for i in range(20):
        i2 = vv.get_info()
        assert i2 == i1
    t2 = time.time()
    # time to get 1st should be much, much longer than subsequent
    assert t2 - t1 < t1 - t0



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Pytest for Compression Model

"""
import pytest
# Import Pyomo libraries
import pyomo.environ as pyo
from pyomo.util.check_units import assert_units_consistent


# Import IDAES core
from idaes.core import FlowsheetBlock
from idaes.core.util.model_statistics import degrees_of_freedom

# Import Unit Model Modules
import idaes.generic_models.properties.swco2 as swco2
from idaes.power_generation.carbon_capture.compression_system.compressor \
      import (CompressionStage, VaneDiffuserType, ImpellerType)

from idaes.core.util.testing import get_default_solver, initialization_tester

# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_default_solver()


@pytest.fixture(scope="module")
def build_unit():
    # Create a Concrete Model as the top level object
    m = pyo.ConcreteModel()
    # Add a flowsheet object to the model
    m.fs = FlowsheetBlock(default={"dynamic": False})
    m.fs.properties_co2 = swco2.SWCO2ParameterBlock(default={
                "phase_presentation": swco2.PhaseType.G
            })
    m.fs.unit = CompressionStage(
        default={"property_package": m.fs.properties_co2,
                 "impeller_type": ImpellerType.open_impeller,
                 "vane_diffuser_type": VaneDiffuserType.vane_diffuser})

    # Set the compressor inlet conditions and an initial flow guess
    p = 1.0951 * 1e5  # Pa
    t = 40.0113 + 273.15  # K #
    fin = 1159.44  # mol/s #

    hin_co2 = swco2.htpx(T=t*pyo.units.K, P=p*pyo.units.Pa)

    m.fs.unit.inlet.flow_mol[:].fix(fin)
    m.fs.unit.inlet.enth_mol[:].fix(hin_co2)
    m.fs.unit.inlet.pressure[:].fix(p)

    # inlet specifications
    m.fs.unit.U2.fix(315.3)
    m.fs.unit.outlet.pressure[:].fix(2.53161*1e5)

    # fix compressor specification
    m.fs.unit.r2.fix(0.67654)
    m.fs.unit.z_s.fix(0.9952)
    m.fs.unit.z_d1.fix(0.97373)
    m.fs.unit.efficiency_mech.fix(0.97)
    m.fs.unit.eff_drive.fix(1.0)

    return m


@pytest.mark.unit
def test_basic_build(build_unit):
    """Make a model and make sure it doesn't throw exception"""
    m = build_unit
    assert degrees_of_freedom(m) == 0
    # Check unit config arguments
    assert len(m.fs.unit.config) == 16
    assert m.fs.unit.config.thermodynamic_assumption
    assert m.fs.unit.config.property_package is m.fs.properties_co2


@pytest.mark.integration
def test_units(build_unit):
    assert_units_consistent(build_unit)


@pytest.mark.component
def test_initialize(build_unit):
    initialization_tester(build_unit, dof=0)


@pytest.mark.component
def test_run(build_unit):
    m = build_unit

    optarg = {"tol": 1e-7,
              "linear_solver": "ma27",
              "max_iter": 50}
    solver.options = optarg
    # solve model
    results = solver.solve(m, tee=True)
    # Check for optimal solution
    assert results.solver.termination_condition == \
        pyo.TerminationCondition.optimal
    assert results.solver.status == pyo.SolverStatus.ok
    assert degrees_of_freedom(m) == 0

    # energy balance
    assert (pytest.approx(0, abs=1e-3) ==
            pyo.value(m.fs.unit.inlet.flow_mol[0]
                      * m.fs.unit.inlet.enth_mol[0]
                      - m.fs.unit.outlet.flow_mol[0]
                      * m.fs.unit.outlet.enth_mol[0]
                      + m.fs.unit.work_mechanical[0]))
    # pressure change
    assert (pytest.approx(143651.0, abs=0.1) ==
            pyo.value(m.fs.unit.deltaP[0]))
    # mass balance
    assert (pytest.approx(0, abs=1e-2) ==
            pyo.value(m.fs.unit.inlet.flow_mol[0]
                      - m.fs.unit.outlet.flow_mol[0]
                      ))



# coding: utf-8
#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
""" __init__.py for idaes module

Set up logging for the idaes module, and import plugins.
"""



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Tests for valve model
"""
import math
import pytest
from pyomo.environ import (ConcreteModel,
                           TerminationCondition,
                           SolverStatus,
                           units,
                           value)
from idaes.core import FlowsheetBlock
from idaes.generic_models.unit_models import Valve, ValveFunctionType
from idaes.core.util.model_statistics import (degrees_of_freedom,
                                              number_variables,
                                              number_total_constraints,
                                              number_unused_variables)
from idaes.core.util.testing import (PhysicalParameterTestBlock,
                                     ReactionParameterTestBlock,
                                     initialization_tester)
from idaes.core.util import get_solver
from pyomo.util.check_units import (assert_units_consistent,
                                    assert_units_equivalent)

from idaes.generic_models.properties import iapws95
import idaes.core.util.scaling as iscale

# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver(options={"nlp_scaling_method": "user-scaling"})


class GenericValve(object):
    @pytest.fixture(scope="class")
    def valve_model(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})
        m.fs.properties = iapws95.Iapws95ParameterBlock()
        m.fs.valve = Valve(default={
            "valve_function_callback":self.type,
            "property_package": m.fs.properties})
        fin = 900 # mol/s
        pin = 200000 # Pa
        pout = 100000 # Pa
        tin = 300 # K
        hin = iapws95.htpx(T=tin*units.K, P=pin*units.Pa) # J/mol
        # Calculate the flow coefficient to give 1000 mol/s flow with given P
        if self.type == ValveFunctionType.linear:
            cv = 1000/math.sqrt(pin - pout)/0.5
        elif self.type == ValveFunctionType.quick_opening:
            cv = 1000/math.sqrt(pin - pout)/math.sqrt(0.5)
        elif self.type == ValveFunctionType.equal_percentage:
            cv = 1000/math.sqrt(pin - pout)/100**(0.5-1)
        # set inlet
        m.fs.valve.inlet.enth_mol[0].fix(hin)
        m.fs.valve.inlet.flow_mol[0].fix(fin)
        m.fs.valve.inlet.flow_mol[0].unfix()
        m.fs.valve.inlet.pressure[0].fix(pin)
        m.fs.valve.outlet.pressure[0].fix(pout)
        m.fs.valve.Cv.fix(cv)
        m.fs.valve.valve_opening.fix(0.5)
        iscale.calculate_scaling_factors(m)
        return m

    @pytest.mark.build
    @pytest.mark.unit
    def test_build(self, valve_model):
        assert hasattr(valve_model.fs.valve, "inlet")
        assert len(valve_model.fs.valve.inlet.vars) == 3
        assert hasattr(valve_model.fs.valve.inlet, "flow_mol")
        assert hasattr(valve_model.fs.valve.inlet, "enth_mol")
        assert hasattr(valve_model.fs.valve.inlet, "pressure")

        assert hasattr(valve_model.fs.valve, "outlet")
        assert len(valve_model.fs.valve.outlet.vars) == 3
        assert hasattr(valve_model.fs.valve.outlet, "flow_mol")
        assert hasattr(valve_model.fs.valve.outlet, "enth_mol")
        assert hasattr(valve_model.fs.valve.outlet, "pressure")

        assert hasattr(valve_model.fs.valve, "pressure_flow_equation")
        assert hasattr(valve_model.fs.valve, "valve_opening")
        assert hasattr(valve_model.fs.valve, "valve_function")
        assert hasattr(valve_model.fs.valve, "flow_var")

        if self.type == ValveFunctionType.equal_percentage:
            # alpha valve function parameter is one more than others
            assert number_variables(valve_model) == 12
        else:
            assert number_variables(valve_model) == 11

        assert number_total_constraints(valve_model) == 6
        assert number_unused_variables(valve_model) == 0

    @pytest.mark.component
    def test_units(self, valve_model):
        assert_units_consistent(valve_model)
        assert_units_equivalent(valve_model.fs.valve.flow_var[0], units.mol/units.s)

    @pytest.mark.unit
    def test_dof(self, valve_model):
        assert degrees_of_freedom(valve_model) == 0

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, valve_model):
        initialization_tester(valve_model, unit=valve_model.fs.valve)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, valve_model):
        results = solver.solve(valve_model)

        # Check for optimal solution
        assert results.solver.termination_condition == \
            TerminationCondition.optimal
        assert results.solver.status == SolverStatus.ok

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, valve_model):
        # calculated Cv to yeild this solution
        assert (pytest.approx(1000, rel=1e-3) ==
                value(valve_model.fs.valve.outlet.flow_mol[0]))

class TestLinearValve(GenericValve):
    type = ValveFunctionType.linear

class TestEPValve(GenericValve):
    type = ValveFunctionType.equal_percentage

class TestQuickValve(GenericValve):
    type = ValveFunctionType.quick_opening



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Flowsheet example of the 2-region bubbling fluidized bed (BFB) model
for a methane combustion with iron-oxide case study

Created: 05/04/2020

Author: Chinedu Okoli
"""
# Import python libraries
import time

# Import Pyomo libraries
from pyomo.environ import ConcreteModel, value

# Import IDAES core modules
from idaes.core import FlowsheetBlock
from idaes.core.util import get_solver

# Import IDAES logger
import idaes.logger as idaeslog

# Import BFB unit model
from idaes.gas_solid_contactors.unit_models.bubbling_fluidized_bed \
    import BubblingFluidizedBed

# Import property packages
from idaes.gas_solid_contactors.properties.methane_iron_OC_reduction. \
    gas_phase_thermo import GasPhaseParameterBlock
from idaes.gas_solid_contactors.properties.methane_iron_OC_reduction. \
    solid_phase_thermo import SolidPhaseParameterBlock
from idaes.gas_solid_contactors.properties.methane_iron_OC_reduction. \
    hetero_reactions import HeteroReactionParameterBlock


# -----------------------------------------------------------------------------
def main():

    # ---------------------------------------------------------------------
    # Build model

    # Create a concrete model
    m = ConcreteModel()

    # Create a steady-state flowsheet
    m.fs = FlowsheetBlock(default={"dynamic": False})

    # Set up thermo-physical and reaction properties
    m.fs.gas_properties = GasPhaseParameterBlock()
    m.fs.solid_properties = SolidPhaseParameterBlock()

    m.fs.hetero_reactions = HeteroReactionParameterBlock(
            default={"solid_property_package": m.fs.solid_properties,
                     "gas_property_package": m.fs.gas_properties})

    # Build the BFB in the flowsheet
    m.fs.BFB = BubblingFluidizedBed(
            default={
                    "flow_type": "co_current",
                    "finite_elements": 5,
                    "transformation_method": "dae.collocation",
                    "gas_phase_config":
                    {"property_package": m.fs.gas_properties},
                    "solid_phase_config":
                    {"property_package": m.fs.solid_properties,
                     "reaction_package": m.fs.hetero_reactions
                     }})

    # ---------------------------------------------------------------------
    # Set design and operating variables of the BFB model

    # Fix design variables
    m.fs.BFB.number_orifice.fix(2500)  # [-]
    m.fs.BFB.bed_diameter.fix(6.5)  # m
    m.fs.BFB.bed_height.fix(5)  # m

    # Fix inlet port variables for gas and solid
    m.fs.BFB.gas_inlet.flow_mol[0].fix(272.81)  # mol/s
    m.fs.BFB.gas_inlet.temperature[0].fix(373)  # K
    m.fs.BFB.gas_inlet.pressure[0].fix(1.86)  # bar
    m.fs.BFB.gas_inlet.mole_frac_comp[0, "CO2"].fix(0.4772)
    m.fs.BFB.gas_inlet.mole_frac_comp[0, "H2O"].fix(0.0646)
    m.fs.BFB.gas_inlet.mole_frac_comp[0, "CH4"].fix(0.4582)

    m.fs.BFB.solid_inlet.flow_mass[0].fix(1230)  # kg/s
    # Particle porosity:
    # The porosity of the OC particle at the inlet is calculated from the
    # known bulk density of the fresh OC particle (3251.75 kg/m3), and the
    # skeletal density of the fresh OC particle (calculated from the known
    # composition of the fresh particle, and the skeletal density of its
    # components [see the solids property package])
    m.fs.BFB.solid_inlet.particle_porosity[0].fix(0.27)
    m.fs.BFB.solid_inlet.temperature[0].fix(1186)  # K
    m.fs.BFB.solid_inlet.mass_frac_comp[0, "Fe2O3"].fix(0.45)
    m.fs.BFB.solid_inlet.mass_frac_comp[0, "Fe3O4"].fix(1e-9)
    m.fs.BFB.solid_inlet.mass_frac_comp[0, "Al2O3"].fix(0.55)

    # ---------------------------------------------------------------------
    # Initialize reactor

    t_start = time.time()  # Run start time

    # State arguments for initializing property state blocks
    # Bubble and gas_emulsion temperatures are initialized at solid
    # temperature because thermal mass of solid >> thermal mass of gas
    blk = m.fs.BFB
    gas_phase_state_args = {
            'flow_mol': blk.gas_inlet.flow_mol[0].value,
            'temperature': blk.solid_inlet.temperature[0].value,
            'pressure': blk.gas_inlet.pressure[0].value,
            'mole_frac': {
                'CH4': blk.gas_inlet.mole_frac_comp[0, 'CH4'].value,
                'CO2': blk.gas_inlet.mole_frac_comp[0, 'CO2'].value,
                'H2O': blk.gas_inlet.mole_frac_comp[0, 'H2O'].value}}
    solid_phase_state_args = {
            'flow_mass': blk.solid_inlet.flow_mass[0].value,
            'particle_porosity': blk.solid_inlet.particle_porosity[0].value,
            'temperature': blk.solid_inlet.temperature[0].value,
            'mass_frac': {
                    'Fe2O3': blk.solid_inlet.mass_frac_comp[0, 'Fe2O3'].value,
                    'Fe3O4': blk.solid_inlet.mass_frac_comp[0, 'Fe3O4'].value,
                    'Al2O3': blk.solid_inlet.mass_frac_comp[0, 'Al2O3'].value}}

    m.fs.BFB.initialize(outlvl=idaeslog.INFO,
                        gas_phase_state_args=gas_phase_state_args,
                        solid_phase_state_args=solid_phase_state_args)

    t_initialize = time.time()  # Initialization time

    # ---------------------------------------------------------------------
    # Final solve

    # Create solver
    solver = get_solver()

    solver.solve(m.fs.BFB, tee=True)

    t_simulation = time.time()  # Simulation time

    print("\n")
    print("----------------------------------------------------------")
    print('Total initialization time: ', value(t_initialize - t_start), " s")
    print("----------------------------------------------------------")

    print("\n")
    print("----------------------------------------------------------")
    print('Total simulation time: ', value(t_simulation - t_start), " s")
    print("----------------------------------------------------------")

    return m


# -----------------------------------------------------------------------------
if __name__ == "__main__":
    m = main()

    # Plot some results
    print()
    stream_table = m.fs.BFB._get_stream_table_contents()
    print(stream_table)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Methods for calculating fugacity of Henry's Law components

For now, only support mole fraction basis
"""
from pyomo.environ import Var


class ConstantH():

    @staticmethod
    def build_parameters(cobj, p):
        b = cobj.parent_block()
        units = b.get_metadata().derived_units

        cobj.add_component(
            "henry_ref_"+p,
            Var(initialize=cobj.config.parameter_data["henry_ref"][p],
                doc="Henry coefficient (mole fraction basis) at reference "
                "state for phase "+p,
                units=units["pressure"]))

    @staticmethod
    def return_expression(b, p, j, T=None):
        cobj = b.params.get_component(j)
        H = getattr(cobj, "henry_ref_"+p)

        return H

    @staticmethod
    def dT_expression(b, p, j, T=None):
        return 0



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
This module contains utilities to provide variable and expression scaling
factors by providing an expression to calculate them via a suffix.

The main purpose of this code is to use the calculate_scaling_factors function
to calculate scaling factors to be used with the Pyomo scaling transformation or
with solvers. A user can provide a scaling_expression suffix to calculate scale
factors from existing variable scaling factors. This allows scaling factors from
a small set of fundamental variables to be propagated to the rest of the model.

The scaling_expression suffix contains Pyomo expressions with model variables.
The expressions can be evaluated with variable scaling factors in place of
variables to calculate additional scaling factors.
"""

__author__ = "John Eslick, Tim Bartholomew, Robert Parker"

from math import log10
import scipy.sparse.linalg as spla
import scipy.linalg as la

import pyomo.environ as pyo
from pyomo.core.expr.visitor import identify_variables
from pyomo.network import Arc
from pyomo.contrib.pynumero.interfaces.pyomo_nlp import PyomoNLP
from pyomo.common.modeling import unique_component_name
from pyomo.core.base.constraint import _ConstraintData
from pyomo.common.collections import ComponentMap
from pyomo.util.calc_var_value import calculate_variable_from_constraint
import idaes.logger as idaeslog

_log = idaeslog.getLogger(__name__)


def __none_left_mult(x, y):
    """PRIVATE FUNCTION, If x is None return None, else return x * y"""
    if x is not None:
        return x * y
    return None


def __scale_constraint(c, v):
    """PRIVATE FUNCTION, scale Constraint c to value v"""
    if c.equality:
        c.set_value((c.lower*v, c.body*v))
    else:
        c.set_value(
            (__none_left_mult(c.lower, v), c.body*v, __none_left_mult(c.upper, v)))


def scale_arc_constraints(blk):
    """Find Arc constraints in a block and its subblocks.  Then scale them based
    on the minimum scaling factor of the variables in the constraint.

    Args:
        blk: Block in which to look for Arc constraints to scale.

    Returns:
        None
    """
    for arc in blk.component_data_objects(Arc, descend_into=True):
        arc_block = arc.expanded_block
        if arc_block is None: # arc not expanded or port empty?
            _log.warning(
                f"{arc} has no constraints. Has the Arc expansion transform "
                "been applied?")
            continue
        for c in arc_block.component_data_objects(pyo.Constraint, descend_into=True):
            sf = min_scaling_factor(identify_variables(c.body))
            constraint_scaling_transform(c, sf)


def map_scaling_factor(iter, default=1, warning=False, func=min, hint=None):
    """Map get_scaling_factor to an iterable of Pyomo components, and call func
    on the result.  This could be use, for example, to get the minimum or
    maximum scaling factor of a set of components.

    Args:
        iter: Iterable yielding Pyomo components
        default: The default value used when a scaling factor is missing. The
            default is default=1.
        warning: Log a warning for missing scaling factors
        func: The function to call on the resulting iterable of scaling factors.
            The default is min().
        hint: Paired with warning=True, this is a string to indicate where the
            missing scaling factor was being accessed, to easier diagnose issues.

    Returns:
        The result of func on the set of scaling factors
    """
    return func(
        map(
            lambda x: get_scaling_factor(
                x, default=default, warning=warning, hint=hint),
            iter
        )
    )


def min_scaling_factor(iter, default=1, warning=True, hint=None):
    """Map get_scaling_factor to an iterable of Pyomo components, and get the
    minimum scaling factor.

    Args:
        iter: Iterable yielding Pyomo components
        default: The default value used when a scaling factor is missing.  If
            None, this will raise an exception when scaling factors are missing.
            The default is default=1.
        warning: Log a warning for missing scaling factors
        hint: Paired with warning=True, this is a string to indicate where the
            missing scaling factor was being accessed, to easier diagnose issues.

    Returns:
        Minimum scaling factor of the components in iter
    """
    return map_scaling_factor(iter, default=default, warning=warning, func=min)


def propagate_indexed_component_scaling_factors(
    blk,
    typ=None,
    overwrite=False,
    descend_into=True):
    """Use the parent component scaling factor to set all component data object
    scaling factors.

    Args:
        blk: The block on which to search for components
        typ: Component type(s) (default=(Var, Constraint, Expression, Param))
        overwrite: if a data object already has a scaling factor should it be
            overwrittten (default=False)
        descend_into: descend into child blocks (default=True)
    """
    if typ is None:
        typ = (pyo.Var, pyo.Constraint, pyo.Expression)

    for c in blk.component_objects(typ, descend_into=descend_into):
        if get_scaling_factor(c) is not None and c.is_indexed():
            for cdat in c.values():
                if overwrite or get_scaling_factor(cdat) is None:
                    set_scaling_factor(cdat, get_scaling_factor(c))


def calculate_scaling_factors(blk):
    """Look for calculate_scaling_factors methods and run them. This uses a
    recursive function to execute the subblock calculate_scaling_factors
    methods first.
    """
    def cs(blk2):
        """ Recursive function for to do subblocks first"""
        for b in blk2.component_data_objects(pyo.Block, descend_into=False):
            cs(b)
        if hasattr(blk2, "calculate_scaling_factors"):
            blk2.calculate_scaling_factors()
    # Call recursive function to run calculate_scaling_factors on blocks from
    # the bottom up.
    cs(blk)
    # If a scale factor is set for an indexed component, propagate it to the
    # component data if a scale factor hasn't already been explicitly set
    propagate_indexed_component_scaling_factors(blk)
    # Use the variable scaling factors to scale the arc constraints.
    scale_arc_constraints(blk)


def set_scaling_factor(c, v, data_objects=True):
    """Set a scaling factor for a model component. This function creates the
    scaling_factor suffix if needed.

    Args:
        c: component to supply scaling factor for
        v: scaling factor
        data_objects: set scaling factors for indexed data objects (default=True)
    Returns:
        None
    """
    if isinstance(c, (float, int)):
        # property packages can return 0 for material balance terms on components
        # doesn't exist.  This handles the case where you get a constant 0 and
        # need it's scale factor to scale the mass balance.
        return 1
    try:
        suf = c.parent_block().scaling_factor
    except AttributeError:
        c.parent_block().scaling_factor = pyo.Suffix(direction=pyo.Suffix.EXPORT)
        suf = c.parent_block().scaling_factor

    suf[c] = v
    if data_objects and c.is_indexed():
        for cdat in c.values():
            suf[cdat] = v


def get_scaling_factor(c, default=None, warning=False, exception=False, hint=None):
    """Get a component scale factor.

    Args:
        c: component
        default: value to return if no scale factor exists (default=None)
        warning: whether to log a warning if a scaling factor is not found
                 (default=False)
        exception: whether to riase an Exception if a scaling factor is not
                   found (default=False)
        hint: (str) a string to add to the warning or exception message to help
            loacate the source.

    Returns:
        scaling factor (float)
    """
    try:
        sf = c.parent_block().scaling_factor[c]
    except (AttributeError, KeyError):
        if hint is None:
            h = ""
        else:
            h = f", {hint}"
        if warning:
            if hasattr(c, "is_component_type") and c.is_component_type():
                _log.warning(f"Missing scaling factor for {c}{h}")
            else:
                _log.warning(f"Trying to get scaling factor for unnamed expr {h}")
        if exception and default is None:
            if hasattr(c, "is_component_type") and c.is_component_type():
                _log.error(f"Missing scaling factor for {c}{h}")
            else:
                _log.error(f"Trying to get scaling factor for unnamed expr {h}")
            raise
        sf = default
    return sf


def unset_scaling_factor(c, data_objects=True):
    """Delete a component scaling factor.

    Args:
        c: component

    Returns:
        None
    """
    try:
        del c.parent_block().scaling_factor[c]
    except (AttributeError, KeyError):
        pass # no scaling factor suffix, is fine
    try:
        if data_objects and c.is_indexed():
            for cdat in c.values():
                del cdat.parent_block().scaling_factor[cdat]
    except (AttributeError, KeyError):
        pass # no scaling factor suffix, is fine


def populate_default_scaling_factors(c):
    """
    Method to set default scaling factors for a number of common quantities
    based of typical values expressed in SI units. Values are converted to
    those used by the property package using Pyomo's unit conversion tools.
    """
    units = c.get_metadata().derived_units

    si_scale = {"temperature": (100*pyo.units.K, "temperature"),
                "pressure": (1e5*pyo.units.Pa, "pressure"),
                "dens_mol_phase": (100*pyo.units.mol/pyo.units.m**3,
                                   "density_mole"),
                "enth_mol": (1e4*pyo.units.J/pyo.units.mol, "energy_mole"),
                "entr_mol": (100*pyo.units.J/pyo.units.mol/pyo.units.K,
                             "entropy_mole"),
                "fug_phase_comp": (1e4*pyo.units.Pa, "pressure"),
                "fug_coeff_phase_comp": (1*pyo.units.dimensionless, None),
                "gibbs_mol": (1e4*pyo.units.J/pyo.units.mol, "energy_mole"),
                "mole_frac_comp": (0.001*pyo.units.dimensionless, None),
                "mole_frac_phase_comp": (0.001*pyo.units.dimensionless, None),
                "mw": (1e-3*pyo.units.kg/pyo.units.mol, "molecular_weight"),
                "mw_phase": (1e-3*pyo.units.kg/pyo.units.mol,
                             "molecular_weight")}

    for p, f in si_scale.items():
        # If a defautl scaling factor exists, do not over write it
        if p not in c.default_scaling_factor.keys():
            if f[1] is not None:
                v = pyo.units.convert(f[0], to_units=units[f[1]])
            else:
                v = f[0]

            sf = 1/(10**round(log10(pyo.value(v))))

            c.set_default_scaling(p, sf)


def __set_constraint_transform_applied_scaling_factor(c, v):
    """PRIVATE FUNCTION Set the scaling factor used to transform a constraint.
    This is used to keep track of scaling transformations that have been applied
    to constraints.

    Args:
        c: component to supply scaling factor for
        v: scaling factor
    Returns:
        None
    """
    try:
        c.parent_block().constraint_transformed_scaling_factor[c] = v
    except AttributeError:
        c.parent_block().constraint_transformed_scaling_factor = pyo.Suffix(
            direction=pyo.Suffix.LOCAL)
        c.parent_block().constraint_transformed_scaling_factor[c] = v


def get_constraint_transform_applied_scaling_factor(c, default=None):
    """Get a the scale factor that was used to transform a
    constraint.

    Args:
        c: constraint data object
        default: value to return if no scaling factor exists (default=None)

    Returns:
        The scaling factor that has been used to transform the constraint or the
        default.
    """
    try:
        sf = c.parent_block().constraint_transformed_scaling_factor.get(c, default)
    except AttributeError:
        sf = default # when there is no suffix
    return sf


def __unset_constraint_transform_applied_scaling_factor(c):
    """PRIVATE FUNCTION: Delete the recorded scale factor that has been used
    to transform constraint c.  This is used when undoing a constraint
    transformation.
    """
    try:
        del c.parent_block().constraint_transformed_scaling_factor[c]
    except AttributeError:
        pass # no scaling factor suffix, is fine
    except KeyError:
        pass # no scaling factor is fine


def constraint_scaling_transform(c, s, overwrite=True):
    """This transforms a constraint by the argument s.  The scaling factor
    applies to original constraint (e.g. if one where to call this twice in a row
    for a constraint with a scaling factor of 2, the original constraint would
    still, only be scaled by a factor of 2.)

    Args:
        c: Pyomo constraint
        s: scale factor applied to the constraint as originally written
        overwrite: overwrite existing scaling factors if present (default=True)

    Returns:
        None
    """
    if not isinstance(c, _ConstraintData):
        raise TypeError(f"{c} is not a constraint or is an indexed constraint")
    st = get_constraint_transform_applied_scaling_factor(c, default=None)

    if not overwrite and st is not None:
        # Existing scaling factor and overwrite False, do nothing
        return

    if st is None:
        # If no existing scaling factor, use value of 1
        st = 1

    v = s/st
    __scale_constraint(c, v)
    __set_constraint_transform_applied_scaling_factor(c, s)


def constraint_scaling_transform_undo(c):
    """The undoes the scaling transforms previously applied to a constraint.

    Args:
        c: Pyomo constraint

    Returns:
        None
    """
    if not isinstance(c, _ConstraintData):
        raise TypeError(f"{c} is not a constraint or is an indexed constraint")
    v = get_constraint_transform_applied_scaling_factor(c)
    if v is None:
        return # hasn't been transformed, so nothing to do.
    __scale_constraint(c, 1/v)
    __unset_constraint_transform_applied_scaling_factor(c)


def unscaled_variables_generator(blk, descend_into=True, include_fixed=False):
    """Generator for unscaled variables

    Args:
        block

    Yields:
        variables with no scale factor
    """
    for v in blk.component_data_objects(pyo.Var, descend_into=descend_into):
        if v.fixed and not include_fixed:
            continue
        if get_scaling_factor(v) is None:
            yield v


def unscaled_constraints_generator(blk, descend_into=True):
    """Generator for unscaled constraints

    Args:
        block

    Yields:
        constraints with no scale factor
    """
    for c in blk.component_data_objects(
        pyo.Constraint, active=True, descend_into=descend_into):
        if get_scaling_factor(c) is None and \
            get_constraint_transform_applied_scaling_factor(c) is None:
            yield c

def constraints_with_scale_factor_generator(blk, descend_into=True):
    """Generator for constraints scaled by a sclaing factor, may or not have
    been transformed.

    Args:
        block

    Yields:
        constraint with a scale factor, scale factor
    """
    for c in blk.component_data_objects(
        pyo.Constraint, active=True, descend_into=descend_into):
        s = get_scaling_factor(c)
        if s is not None:
            yield c, s


def badly_scaled_var_generator(
    blk, large=1e4, small=1e-3, zero=1e-10, descend_into=True, include_fixed=False):
    """This provides a rough check for variables with poor scaling based on
    their current scale factors and values. For each potentially poorly scaled
    variable it returns the var and its current scaled value.

    Args:
        blk: pyomo block
        large: Magnitude that is considered to be too large
        small: Magnitude that is considered to be too small
        zero: Magnitude that is considered to be zero, variables with a value of
            zero are okay, and not reported.

    Yields:
        variable data object, current absolute value of scaled value
    """
    for v in blk.component_data_objects(pyo.Var, descend_into=descend_into):
        if v.fixed and not include_fixed:
            continue
        val = pyo.value(v, exception=False)
        if val is None:
            continue
        sf = get_scaling_factor(v, default=1)
        sv = abs(val * sf)  # scaled value
        if sv > large:
            yield v, sv
        elif sv < zero:
            continue
        elif sv < small:
            yield v, sv


def constraint_autoscale_large_jac(
    m,
    ignore_constraint_scaling=False,
    ignore_variable_scaling=False,
    max_grad=100,
    min_scale=1e-6,
    no_scale=False
):
    """Automatically scale constraints based on the Jacobian.  This function
    imitates Ipopt's default constraint scaling.  This scales constraints down
    to avoid extremely large values in the Jacobian.  This function also returns
    the unscaled and scaled Jacobian matrixes and the Pynumero NLP which can be
    used to identify the constraints and variables corresponding to the rows and
    comlumns.

    Args:
        m: model to scale
        ignore_constraint_scaling: ignore existing constraint scaling
        ignore_variable_scaling: ignore existing variable scaling
        max_grad: maximum value in Jacobian after scaling, subject to minimum
            scaling factor restriction.
        min_scale: minimum scaling factor allowed, keeps constraints from being
            scaled too much.
        no_scale: just calculate the Jacobian and scaled Jacobian, don't scale
            anything

    Returns:
        unscaled Jacobian CSR from, scaled Jacobian CSR from, Pynumero NLP
    """
    # Pynumero requires an objective, but I don't, so let's see if we have one
    n_obj = 0
    for c in m.component_data_objects(pyo.Objective, active=True):
        n_obj += 1
    # Add an objective if there isn't one
    if n_obj == 0:
        dummy_objective_name = unique_component_name(m, "objective")
        setattr(m, dummy_objective_name, pyo.Objective(expr=0))
    # Create NLP and calculate the objective
    nlp = PyomoNLP(m)
    jac = nlp.evaluate_jacobian().tocsr()
    # Get lists of varibles and constraints to translate Jacobian indexes
    # save them on the NLP for later, since genrating them seems to take a while
    nlp.clist = clist = nlp.get_pyomo_constraints()
    nlp.vlist = vlist = nlp.get_pyomo_variables()
    # Create a scaled Jacobian to account for variable scaling, for now ignore
    # constraint scaling
    jac_scaled = jac.copy()
    for i, c in enumerate(clist):
        for j in jac_scaled[i].indices:
            v = vlist[j]
            if ignore_variable_scaling:
                sv = 1
            else:
                sv = get_scaling_factor(v, default=1)
            jac_scaled[i,j] = jac_scaled[i,j]/sv
    # calculate constraint scale factors
    for i, c in enumerate(clist):
        sc = get_scaling_factor(c, default=1)
        if not no_scale:
            if (ignore_constraint_scaling or get_scaling_factor(c) is None):
                row = jac_scaled[i]
                for d in row.indices:
                    row[0,d] = abs(row[0,d])
                mg = row.max()
                if mg > max_grad:
                    sc = max(min_scale, max_grad/mg)
                set_scaling_factor(c, sc)
        for j in jac_scaled[i].indices:
            # update the scaled jacobian
            jac_scaled[i,j] = jac_scaled[i,j]*sc
    # delete dummy objective
    if n_obj == 0:
        delattr(m, dummy_objective_name)
    return jac, jac_scaled, nlp


def get_jacobian(m, scaled=True):
    """
    Get the Jacobian matrix at the current model values. This function also
    returns the Pynumero NLP which can be used to identify the constraints and
    variables corresponding to the rows and comlumns.

    Args:
        m: model to get Jacobian from
        scaled: if True return scaled Jacobian, else get unscaled

    Returns:
        Jacobian matrix in Scipy CSR format, Pynumero nlp
    """
    jac, jac_scaled, nlp = constraint_autoscale_large_jac(m, no_scale=True)
    if scaled:
        return jac_scaled, nlp
    else:
        return jac, nlp


def extreme_jacobian_entries(
        m=None, scaled=True, large=1e4, small=1e-4, zero=1e-10, jac=None, nlp=None):
    """
    Show very large and very small Jacobian entries.

    Args:
        m: model
        scaled: if true use scaled Jacobian
        large: >= to this value is consdered large
        small: <= to this and >= zero is consdered small

    Returns:
        (list of tuples), Jacobian entry, Constraint, Variable
    """
    if jac is None or nlp is None:
        jac, nlp = get_jacobian(m, scaled)
    el = []
    for i, c in enumerate(nlp.clist):
        for j in jac[i].indices:
            v = nlp.vlist[j]
            e = abs(jac[i, j])
            if (e <= small and e > zero) or e >= large:
                el.append((e, c, v))
    return el


def jacobian_cond(m=None, scaled=True, ord=None, pinv=False, jac=None):
    """
    Get the condition number of the scaled or unscaled Jacobian matrix of a model.

    Args:
        m: calculate the condition number of the Jacobian from this model.
        scaled: if True use scaled Jacobian, else use unscaled
        ord: norm order, None = Frobenius, see scipy.sparse.linalg.norm for more
        pinv: Use pseudoinverse, works for non-square matrixes
        jac: (optional) perviously calculated jacobian

    Returns:
        (float) Condition number
    """
    if jac is None:
        jac, nlp = get_jacobian(m, scaled)
    jac = jac.tocsc()
    if jac.shape[0] != jac.shape[1] and not pinv:
        _log.warning("Nonsquare Jacobian using pseudo inverse")
        pinv = True
    if not pinv:
        jac_inv = spla.inv(jac)
        return spla.norm(jac, ord)*spla.norm(jac_inv, ord)
    else:
        jac_inv = la.pinv(jac.toarray())
        return spla.norm(jac, ord)*la.norm(jac_inv, ord)


class CacheVars(object):
    """
    A class for saving the values of variables then reloading them,
    usually after they have been used to perform some solve or calculation.
    """
    def __init__(self, vardata_list):
        self.vars = vardata_list
        self.cache = [None for var in self.vars]

    def __enter__(self):
        for i, var in enumerate(self.vars):
            self.cache[i] = var.value
        return self

    def __exit__(self, ex_type, ex_value, ex_traceback):
        for i, var in enumerate(self.vars):
            var.set_value(self.cache[i])


class FlattenedScalingAssignment(object):
    """
    A class to assist in the calculation of scaling factors when a
    variable-constraint assignment can be constructed, especially when
    the variables and constraints are all indexed by some common set(s).
    """
    def __init__(self, scaling_factor, varconlist=None, nominal_index=()):
        """
        Args:
            scaling_factor: A Pyomo scaling_factor Suffix that will hold all
                            the scaling factors calculated
            varconlist: A list of variable, constraint tuples. These variables
                        and constraints should be indexed by the same sets,
                        so they may need to be references-to-slices along some
                        common sets.
            nominal_index: The index of variables and constraints to access
                           when a calculation needs to be performed using
                           data objects.
        """
        if varconlist is None:
            varconlist = []

        self.scaling_factor = scaling_factor
        self.nominal_index = nominal_index
        if nominal_index is None or nominal_index == ():
            self.dim = 0
        else:
            try:
                self.dim = len(nominal_index)
            except TypeError:
                self.dim = 1

        varlist = []
        conlist = []
        for var, con in varconlist:
            varlist.append(var)
            conlist.append(con)
        self.varlist = varlist
        self.conlist = conlist

        data_getter = self.get_representative_data_object
        var_con_data_list = [(data_getter(var), data_getter(con))
                for var, con in varconlist]
        con_var_data_list = [(data_getter(con), data_getter(var))
                for var, con in varconlist]
        self.var2con = ComponentMap(var_con_data_list)
        self.con2var = ComponentMap(con_var_data_list)

    def get_representative_data_object(self, obj):
        """
        Gets a data object from an object of the appropriate dimension
        """
        if self.dim == 0:
            # In this way, obj can be a data object and this class can be
            # used even if the assignment is not between "flattened components"
            return obj
        else:
            nominal_index = self.nominal_index
            return obj[nominal_index]

    def calculate_variable_scaling_factor(self, var, include_fixed=False):
        """
        Calculates the scaling factor of a variable based on the
        constraint assigned to it. Loads each variable in that constraint
        with its nominal value (inverse of scaling factor), calculates
        the value of the target variable from the constraint, then sets
        its scaling factor to the inverse of the calculated value.
        """
        vardata = self.get_representative_data_object(var)
        condata = self.var2con[vardata]
        scaling_factor = self.scaling_factor

        in_constraint = list(identify_variables(
            condata.expr,
            include_fixed=include_fixed,
            ))
        source_vars = [v for v in in_constraint if v is not vardata]
        nominal_source = [1/scaling_factor[var] for var in source_vars]

        with CacheVars(in_constraint) as cache:
            for v, nom_val in zip(source_vars, nominal_source):
                v.set_value(nom_val)
            # This assumes that target var is initialized to a somewhat
            # reasonable value
            calculate_variable_from_constraint(vardata, condata)
            nominal_target = vardata.value
        if nominal_target == 0:
            target_factor = 1.0
        else:
            target_factor = abs(1/nominal_target)

        if self.dim == 0:
            scaling_factor[var] = target_factor
        else:
            for v in var.values():
                scaling_factor[v] = target_factor

    def set_constraint_scaling_factor(self, con):
        """
        Sets the scaling factor of a constraint to that of its assigned variable
        """
        condata = self.get_representative_data_object(con)
        vardata = self.con2var[condata]
        scaling_factor = self.scaling_factor

        var_factor = scaling_factor[vardata]
        if self.dim == 0:
            scaling_factor[con] = var_factor
        else:
            for c in con.values():
                scaling_factor[c] = var_factor

    def set_derivative_factor_from_state(self, deriv, nominal_wrt=1.0):
        """
        Sets the scaling factor for a DerivativeVar equal to the factor for
        its state var at every index. This method needs access to the
        get_state_var method, so deriv must be an actual DerivativeVar,
        not a reference-to-slice.
        """
        scaling_factor = self.scaling_factor
        state_var = deriv.get_state_var()
        for index, dv in deriv.items():
            state_data = state_var[index]
            nominal_state = 1/scaling_factor[state_data]
            nominal_deriv = nominal_state/nominal_wrt
            scaling_factor[dv] = 1/nominal_deriv


################################################################################
# DEPRECATED functions below.
################################################################################

def scale_single_constraint(c):
    """This transforms a constraint with its scaling factor. If there is no
    scaling factor for the constraint, the constraint is not scaled and a
    message is logged. After transforming the constraint the scaling factor,
    scaling expression, and nominal value are all unset to ensure the constraint
    isn't scaled twice.

    Args:
        c: Pyomo constraint

    Returns:
        None
    """
    _log.warning(
        "DEPRECATED: scale_single_constraint() will be removed and has no "
        "direct replacement")
    if not isinstance(c, _ConstraintData):
        raise TypeError(
            "{} is not a constraint and cannot be the input to "
            "scale_single_constraint".format(c.name))

    v = get_scaling_factor(c)
    if v is None:
        _log.warning(
            f"{c.name} constraint has no scaling factor, so it was not scaled.")
        return
    __scale_constraint(c, v)
    unset_scaling_factor(c)


def scale_constraints(blk, descend_into=True):
    """This scales all constraints with their scaling factor suffix for a model
    or block. After scaling the constraints, the scaling factor and expression
    for each constraint is set to 1 to avoid double scaling the constraints.

    Args:
        blk: Pyomo block
        descend_into: indicates whether to descend into the other blocks on blk.
            (default = True)

    Returns:
        None
    """
    _log.warning(
        "DEPRECATED: scale_single_constraint() will be removed and has no "
        "direct replacement")
    for c in blk.component_data_objects(pyo.Constraint, descend_into=False):
        scale_single_constraint(c)
    if descend_into:
        for b in blk.component_objects(pyo.Block, descend_into=True):
            for c in b.component_data_objects(pyo.Constraint, descend_into=False):
                scale_single_constraint(c)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Tests for smooth VLE formulation

Authors: Andrew Lee
"""

import pytest
from sys import modules

from pyomo.environ import (ConcreteModel,
                           Constraint,
                           Set,
                           value,
                           Var,
                           units as pyunits)

from idaes.generic_models.properties.core.phase_equil.bubble_dew import \
    IdealBubbleDew
from idaes.core import declare_process_block_class
from idaes.generic_models.properties.core.generic.generic_property import (
        GenericParameterData)
from idaes.generic_models.properties.core.generic.tests.dummy_eos import DummyEoS


# Dummy class to use for Psat calls
Psat = {"H2O": 1e5, "EtOH": 5e4}


# Dummy method to avoid errors when setting metadata dict
def set_metadata(b):
    pass


def pressure_sat_comp(b, j, T):
    return Psat[j.local_name]


@declare_process_block_class("DummyParameterBlock")
class DummyParameterData(GenericParameterData):
    def configure(self):
        self.configured = True

    def parameters(self):
        self.parameters_set = True


def define_state(b):
    b.state_defined = True


@pytest.fixture(scope="class")
def frame():
    m = ConcreteModel()

    # Dummy params block
    m.params = DummyParameterBlock(default={
                "components": {
                    "H2O": {"pressure_sat_comp": pressure_sat_comp},
                    "EtOH": {"pressure_sat_comp": pressure_sat_comp}},
                "phases": {"Liq": {"equation_of_state": DummyEoS},
                           "Vap": {"equation_of_state": DummyEoS}},
                "state_definition": modules[__name__],
                "pressure_ref": 1e5,
                "temperature_ref": 300,
                "base_units": {"time": pyunits.s,
                               "length": pyunits.m,
                               "mass": pyunits.kg,
                               "amount": pyunits.mol,
                               "temperature": pyunits.K}})
    m.params._pe_pairs = Set(initialize=[("Vap", "Liq")])

    m.props = m.params.build_state_block([1],
                                         default={"defined_state": False})

    # Add common variables
    m.props[1].pressure = Var(initialize=101325)
    m.props[1].temperature = Var(initialize=300)
    m.props[1]._teq = Var(initialize=300)
    m.props[1].mole_frac_comp = Var(m.params.component_list,
                                    initialize=0.5)

    return m


class TestBubbleTempIdeal(object):
    @pytest.mark.unit
    def test_build(self, frame):
        frame.props[1].temperature_bubble = Var(frame.params._pe_pairs)
        frame.props[1]._mole_frac_tbub = Var(frame.params._pe_pairs,
                                             frame.params.component_list,
                                             initialize=0.5)

        IdealBubbleDew.temperature_bubble(frame.props[1])

        assert isinstance(frame.props[1].eq_temperature_bubble, Constraint)
        assert len(frame.props[1].eq_temperature_bubble) == 1

        assert isinstance(frame.props[1].eq_mole_frac_tbub, Constraint)
        assert len(frame.props[1].eq_mole_frac_tbub) == 2
        for k in frame.props[1].eq_mole_frac_tbub:
            assert k in [("Vap", "Liq", "H2O"), ("Vap", "Liq", "EtOH")]

    @pytest.mark.unit
    def test_expressions(self, frame):
        for x1 in range(0, 11, 1):
            frame.props[1].mole_frac_comp["H2O"].value = x1/10
            frame.props[1].mole_frac_comp["EtOH"].value = 1 - x1/10

            frame.props[1].pressure = value(
                sum(frame.props[1].mole_frac_comp[j] * Psat[j]
                    for j in frame.params.component_list))

            for pp in frame.params._pe_pairs:
                for j in frame.params.component_list:
                    frame.props[1]._mole_frac_tbub[pp[0], pp[1], j].value = \
                        value(frame.props[1].mole_frac_comp[j] * Psat[j] /
                              frame.props[1].pressure)

                assert value(
                    frame.props[1].eq_temperature_bubble[
                        pp[0], pp[1]].body) == pytest.approx(0, abs=1e-8)
                for k in frame.params.component_list:
                    assert value(frame.props[1].eq_mole_frac_tbub[
                        pp[0], pp[1], k].body) == pytest.approx(0, abs=1e-8)


class TestDewTempIdeal(object):
    @pytest.mark.unit
    def test_build(self, frame):
        frame.props[1].temperature_dew = Var(frame.params._pe_pairs)
        frame.props[1]._mole_frac_tdew = Var(frame.params._pe_pairs,
                                             frame.params.component_list,
                                             initialize=0.5)

        IdealBubbleDew.temperature_dew(frame.props[1])

        assert isinstance(frame.props[1].eq_temperature_dew, Constraint)
        assert len(frame.props[1].eq_temperature_dew) == 1

        assert isinstance(frame.props[1].eq_mole_frac_tdew, Constraint)
        assert len(frame.props[1].eq_mole_frac_tdew) == 2
        for k in frame.props[1].eq_mole_frac_tdew:
            assert k in [("Vap", "Liq", "H2O"), ("Vap", "Liq", "EtOH")]

    @pytest.mark.unit
    def test_expressions(self, frame):
        for x1 in range(0, 11, 1):
            frame.props[1].mole_frac_comp["H2O"].value = x1/10
            frame.props[1].mole_frac_comp["EtOH"].value = 1 - x1/10

            frame.props[1].pressure = value(
                1 / sum(frame.props[1].mole_frac_comp[j] / Psat[j]
                        for j in frame.params.component_list))

            for pp in frame.params._pe_pairs:
                for j in frame.params.component_list:
                    frame.props[1]._mole_frac_tdew[pp[0], pp[1], j].value = \
                        value(frame.props[1].mole_frac_comp[j] *
                              frame.props[1].pressure / Psat[j])

                assert value(frame.props[1].eq_temperature_dew[
                    pp[0], pp[1]].body) == pytest.approx(0, abs=1e-8)
                for k in frame.params.component_list:
                    assert value(frame.props[1].eq_mole_frac_tdew[
                        pp[0], pp[1], k].body) == pytest.approx(0, abs=1e-8)


class TestBubblePresIdeal(object):
    @pytest.mark.unit
    def test_build(self, frame):
        frame.props[1].pressure_bubble = Var(frame.params._pe_pairs)
        frame.props[1]._mole_frac_pbub = Var(frame.params._pe_pairs,
                                             frame.params.component_list,
                                             initialize=0.5)

        IdealBubbleDew.pressure_bubble(frame.props[1])

        assert isinstance(frame.props[1].eq_pressure_bubble, Constraint)
        assert len(frame.props[1].eq_pressure_bubble) == 1

        assert isinstance(frame.props[1].eq_mole_frac_pbub, Constraint)
        assert len(frame.props[1].eq_mole_frac_pbub) == 2
        for k in frame.props[1].eq_mole_frac_pbub:
            assert k in [("Vap", "Liq", "H2O"), ("Vap", "Liq", "EtOH")]

    @pytest.mark.unit
    def test_expressions(self, frame):
        for x1 in range(0, 11, 1):
            frame.props[1].mole_frac_comp["H2O"].value = x1/10
            frame.props[1].mole_frac_comp["EtOH"].value = 1 - x1/10

            for pp in frame.params._pe_pairs:
                frame.props[1].pressure_bubble[pp[0], pp[1]] = value(
                    sum(frame.props[1].mole_frac_comp[j] * Psat[j]
                        for j in frame.params.component_list))

                for j in frame.params.component_list:
                    frame.props[1]._mole_frac_pbub[pp[0], pp[1], j].value = \
                        value(frame.props[1].mole_frac_comp[j] * Psat[j] /
                              frame.props[1].pressure_bubble[pp[0], pp[1]])

                assert value(frame.props[1].eq_pressure_bubble[
                    pp[0], pp[1]].body) == pytest.approx(0, abs=1e-8)
                for k in frame.params.component_list:
                    assert value(frame.props[1].eq_mole_frac_pbub[
                        pp[0], pp[1], k].body) == pytest.approx(0, abs=1e-8)


class TestDewPressureIdeal(object):
    @pytest.mark.unit
    def test_build(self, frame):
        frame.props[1].pressure_dew = Var(frame.params._pe_pairs)
        frame.props[1]._mole_frac_pdew = Var(frame.params._pe_pairs,
                                             frame.params.component_list,
                                             initialize=0.5)

        IdealBubbleDew.pressure_dew(frame.props[1])

        assert isinstance(frame.props[1].eq_pressure_dew, Constraint)
        assert len(frame.props[1].eq_pressure_dew) == 1

        assert isinstance(frame.props[1].eq_mole_frac_pdew, Constraint)
        assert len(frame.props[1].eq_mole_frac_pdew) == 2
        for k in frame.props[1].eq_mole_frac_pdew:
            assert k in [("Vap", "Liq", "H2O"), ("Vap", "Liq", "EtOH")]

    @pytest.mark.unit
    def test_expressions(self, frame):
        for x1 in range(0, 11, 1):
            frame.props[1].mole_frac_comp["H2O"].value = x1/10
            frame.props[1].mole_frac_comp["EtOH"].value = 1 - x1/10

            for pp in frame.params._pe_pairs:
                frame.props[1].pressure_dew[pp[0], pp[1]] = value(
                    1 / sum(frame.props[1].mole_frac_comp[j] / Psat[j]
                            for j in frame.params.component_list))

                for j in frame.params.component_list:
                    frame.props[1]._mole_frac_pdew[pp[0], pp[1], j].value = \
                        value(frame.props[1].mole_frac_comp[j] *
                              frame.props[1].pressure_dew[pp[0], pp[1]] /
                              Psat[j])

                assert value(frame.props[1].eq_pressure_dew[
                    pp[0], pp[1]].body) == pytest.approx(0, abs=1e-8)
                for k in frame.params.component_list:
                    assert value(frame.props[1].eq_mole_frac_pdew[
                        pp[0], pp[1], k].body) == pytest.approx(0, abs=1e-8)
